import numpy as np
import dcmri.pk as pk
import dcmri.lib as lib


def aif_tristan(
        t:np.ndarray, 
        agent = 'gadoterate',
        dose = 0.2, 
        rate = 3, 
        BAT = 0, 
        weight = 73, 
        CO = 97, 
        E = 0.07, 
        Thl = 13,  
        Dhl = 0.5,  
        Tp = 25, 
        Te = 350, 
        Ee = 0.18, 
        dtol = 0.01, 
    )->np.ndarray:
    """Arterial input function with default parameters for young healthy volunteers.

    This AIF was measured in the TRISTAN project (Min et al 2024). The default values are for young healthy volunteers but since the AIF is built on a whole-body model of the circulation, they can be modified to generate virtual populations. 

    Reference:
    
    Thazin Min, Marta Tibiletti, Paul Hockings, Aleksandra Galetin, Ebony Gunwhy, Gerry Kenna, Nicola Melillo, Geoff JM Parker, Gunnar Schuetz, Daniel Scotcher, John Waterton, Ian Rowe, and Steven Sourbron. *Measurement of liver function with dynamic gadoxetate-enhanced MRI: a validation study in healthy volunteers*. Proc Intl Soc Mag Reson Med, Singapore 2024.

    Args:
        t (np.ndarray): Array of time points
        agent (str, optional): Contrast agent generic name. Defaults to 'gadoterate'.
        dose (float, optional): Contrast agent dose in mL/kg. Defaults to 0.2.
        rate (float, optional): Contrast agent injection rate in mL/sec. Defaults to 3.
        BAT (float, optional): Bolus arrival time in sec. Defaults to 0.
        weight (float, optional): Subject weight in kg. Defaults to 73.
        CO (float, optional): Cardiac output in mL/sec. Defaults to 97.
        E (float, optional): Body extraction fraction. Defaults to 0.07.
        Thl (float, optional): Mean transit time of the heart-lung system in sec. Defaults to 13.
        Dhl (float, optional): Transit time dispersion of the heart-lung systen. Defaults to 0.5.
        Tp (float, optional): Plasma mean transit time in sec of the other organs. Defaults to 25.
        Te (float, optional): Extravascular mean transit time in sec. Defaults to 350.
        Ee (float, optional): Extraction fraction into the extravascular space. Defaults to 0.18.
        dtol (float, optional): Dose tolerance. Defaults to 0.01.

    Returns:
        np.ndarray: Aorta concentrations in mmol/mL.

    Example:

        Generate AIFs with different levels of cardiac output:

    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        import dcmri as dc

        # Time points in sec
        t = np.arange(0, 180, 2.0)

        # Plot aifs with different levels of cardiac output, including the default of 100mL/sec.
        plt.plot(t/60, 1000*dc.aif_tristan(t, BAT=30), 'r-', label='AIF (default)')
        plt.plot(t/60, 1000*dc.aif_tristan(t, BAT=30, CO=150), 'g-', label='AIF (increased cardiac output)')
        plt.plot(t/60, 1000*dc.aif_tristan(t, BAT=30, CO=75), 'b-', label='AIF (reduced cardiac output)')
        plt.xlabel('Time (min)')
        plt.ylabel('Concentration (mmol/mL)')
        plt.legend()
        plt.show()    
    """
    conc = lib.ca_conc(agent)
    Ji = lib.influx_step(t, weight, 
            conc, dose, rate, BAT,
    )
    Jb = flux_aorta(Ji, t, E=E,
            heartlung = ['chain', (Thl, Dhl)], 
            organs=['2cxm', ([Tp, Te], Ee)],
            tol=dtol,   
    )
    return Jb/CO



def flux_aorta(J_vena:np.ndarray, 
        t=None, dt=1.0, E=0.1, FFkl=0.0, FFk=0.5, 
        heartlung = ['pfcomp', (10, 0.2)],
        organs = ['2cxm', ([20, 120], 0.15)],
        kidneys = ['comp', (10,)],
        liver = ['pfcomp', (10, 0.2)],
        tol = 0.001):
    
    """Indicator flux through the body modelled with a two-system whole body model (heart-lung system and organ system).

    The heart-lung system and the other organs are organised in a loop, with the heart and lungs modelled as a chain and the other organs as a 2-compartment exchange model (`flux_2comp`).

    Args:
        J_vena (np.ndarray): Indicator influx (mmol/sec) into the veins. 
        heartlung (list): 3-element list specifying the model to use for the heart-lung system (see notes for detail).
        organs (dict): 3-element list specifying the model to use for the organs (see notes for detail). 
        kidneys (dict): 3-element list specifying the model to use for the kidneys (see notes for detail). 
        liver (dict): 3-element list specifying the model to use for the liver (see notes for detail). 
        t (np.ndarray, optional): Array of time points (sec), must be of equal size as J_vena. If not provided, the time points are uniformly sampled with interval dt. Defaults to None.
        dt (float, optional): Sampling interval in sec. Defaults to 1.0.
        tol (float, optional): Dose tolerance in the solution. The solution propagates the input through the system, until the dose that is left in the system is given by tol*dose0, where dose0 is the initial dose. Defaults to 0.001.

    Returns:
        tuple: Indicator fluxes (mmol/sec) through the vena cava and aorta.

    Notes:

        1. The lists specifying each organ system consist of 3 elements: the model (str), it parameters (tuple) and any keyword parameters (dict). Any of the basic pk blocks can be used, for instance *chain*, *plug-flow compartment*, and *compartment* would be specified as follows:
        - **chain**: ['chain', (Thl, Dhl), {'solver':'step'}]
        - **plugflow-compartment**: ['pfcomp', (Thl, Dhl), {'solver':'interp'}}
        - **compartment**: ['comp', Thl]
        - **2cxm**: ['2cxm', ([To, Te], Eo)]
        - and so on..

        2. Different architectures can be modelled by setting the flow fractions:
        - Set FFkl=0 to model organs only
        - Set FFk=1 to model organs+kidney
        - Set FFk=0 to model organs+liver.


    Example:

        Generate flux through aorta:

    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        import dcmri as dc

        # Generate a stepwise injection:
        t = np.arange(0, 120, 2.0)
        Ji = dc.influx_step(t, 70, 0.5, 0.2, 3, 30)

        # Calculate the fluxes in mmol/sec:
        Ja = dc.flux_aorta(Ji, t)

        # Plot the fluxes:
        plt.plot(t/60, Ja, 'r-', label='Aorta')
        plt.xlabel('Time (min)')
        plt.ylabel('Indicator flux (mmol/sec)')
        plt.legend()
        plt.show()
    """

    dose = np.trapz(J_vena, x=t, dx=dt)
    min_dose = tol*dose

    # Residuals of each pathway
    Rk = FFk*FFkl*(1-E)
    Rl = (1-FFk)*FFkl*(1-E)
    Ro = (1-FFkl)*(1-E)

    # Initialize output
    J_aorta_total = np.zeros(J_vena.size)

    while dose > min_dose:

        # Aorta flux of the current pass
        J_aorta = pk.flux(J_vena, *heartlung[1], t=t, dt=dt, kinetics=heartlung[0])

        # Add to the total aorta flux
        J_aorta_total += J_aorta

        # Venous flux of the current pass
        J_vena = Ro*pk.flux(J_aorta, *organs[1], t=t, dt=dt, kinetics=organs[0])
        if Rl>0:
            J_vena += Rl*pk.flux(J_aorta, *liver[1], t=t, dt=dt, kinetics=liver[0])
        if Rk>0:
            J_vena += Rk*pk.flux(J_aorta, *kidneys[1], t=t, dt=dt, kinetics=kidneys[0])

        # Get residual dose in current pass
        dose = np.trapz(J_vena, x=t, dx=dt)

    return J_aorta_total




import matplotlib.pyplot as plt
import numpy as np
import dcmri as dc


class Liver(dc.Model):

    """Single-inlet model for intracellular indicator measured with a steady-state sequence.

        **Input function**

        - **aif** (array-like, default=None): Signal-time curve in a feeding artery. If AIF is set to None, then the parameter ca must be provided (arterial concentrations).
        - **ca** (array-like, default=None): Concentration (M) in the arterial input. Must be provided when aif = None, ignored otherwise.

        **Acquisition parameters**

        - **sequence** (str, default='SS'): imaging sequence.
        - **t** (array-like, default=None): array of time points.
        - **dt** (float, optional): Sampling interval of the AIF in sec. 
        - **agent** (str, optional): Contrast agent generic name.
        - **field_strength** (float, optional): Magnetic field strength in T. 
        - **TR** (float, optional): Repetition time, or time between excitation pulses, in sec. 
        - **FA** (float, optional): Nominal flip angle in degrees.
        - **n0** (float, optional): Baseline length in number of acquisitions.

        **Kinetic parameters**

        - **kinetics** (str, default='stationary'): Tracer-kinetic model.
        - **Hct** (float, optional): Hematocrit. 
        - **Te**: mean transit time of the extracellular space.
        - **De**: Transit time dispersion of the extracellular space, in the range [0,1].
        - **ve**: volume faction of the extracellular space.
        - **khe**: rate constant for indicator transport from extracellular space to hepatocytes, in mL/sec/mL. 
        - **Th**: mean transit time of the hepatocytes (sec).

        **Prediction and training parameters**

        - **free** (array-like): list of free parameters. The default depends on the kinetics parameter.
        - **bounds** (array-like): 2-element list with lower and upper bounds of the free parameters. The default depends on the kinetics parameter.

        **Signal parameters**

        - **R10** (float, optional): Precontrast tissue relaxation rate in 1/sec. 
        - **S0** (float, optional): Signal scaling factor.

        **Other parameters**

        - **vol** (float, optional): Liver volume in mL.

    Args:
        params (dict, optional): override defaults for any of the parameters.

    See Also:
        `Tissue`

    Example:

        Derive model parameters from simulated data:

    .. plot::
        :include-source:
        :context: close-figs
    
        >>> import matplotlib.pyplot as plt
        >>> import dcmri as dc

        Use `fake_tissue` to generate synthetic test data:

        >>> time, aif, roi, gt = dc.fake_tissue(CNR=100, agent='gadoxetate', R10=1/dc.T1(3.0,'liver'))
        
        Build a tissue model and set the constants to match the experimental conditions of the synthetic test data:

        >>> model = dc.Liver(
        ...     aif = aif,
        ...     dt = time[1],
        ...     Hct = 0.45, 
        ...     agent = 'gadoxetate',
        ...     TR = 0.005,
        ...     FA = 20,
        ...     n0 = 10,
        ... )

        Train the model on the ROI data:

        >>> model.train(time, roi)

        Plot the reconstructed signals (left) and concentrations (right) and compare the concentrations against the noise-free ground truth:

        >>> model.plot(time, roi, testdata=gt)
    """ 

    def __init__(self, **params):

        # Input function
        self.aif = None
        self.ca = None

        # Acquisition parameters
        self.sequence = 'SS'
        self.field_strength = 3.0 
        self.t = None
        self.dt = 0.5                    
        self.agent = 'gadoxetate'  
        self.n0 = 1  
        self.TR = 0.005  
        self.FA = 15.0  
        self.TC = 0.180

        # Tracer-kinetic parameters
        self.kinetics = 'non-stationary'
        self.Hct = 0.45    
        self.Te = 30.0
        self.De = 0.85
        self.ve = 0.3 
        self.khe = 20/6000
        self.Th = 30*60
        self.khe_f = 20/6000
        self.Th_f = 30*60

        # Signal parameters
        self.R10 = 1/dc.T1(3.0, 'liver')  
        self.R10b = 1/dc.T1(3.0, 'blood')  
        self.S0 = 1         
        
        # training parameters
        self.free = ['Te','De','ve','khe','Th']
        self.bounds = [
            [0.1, 0, 0.01, 0, 10*60, ],
            [60, 1, 0.6, 0.1, 10*60*60],
        ]

        # Other parameters
        self.vol = None
        
        # Preset parameters
        if 'kinetics' in params:
            if params['kinetics'] == 'non-stationary':
                self.free += ['khe_f','Th_f']
                self.bounds[0] += [0, 10*60]
                self.bounds[1] += [ 0.1, 10*60*60]
        
        # Override defaults
        for k, v in params.items():
            setattr(self, k, v)

        # Check inputs
        if (self.aif is None) and (self.ca is None):
            raise ValueError('Please provide either an arterial sigal (aif) or an arterial concentration (ca).')  
        
    def time(self):
        """Array of time points.

        Returns:
            np.ndarray: time points in seconds.
        """
        if self.t is None:
            if self.aif is None:
                return self.dt*np.arange(np.size(self.ca))
            else:
                return self.dt*np.arange(np.size(self.aif))
        else:
            return self.t


    def conc(self, sum=True):
        """Tissue concentrations

        Args:
            sum (bool, optional): If True, returns the total concentrations. If False, returns concentration in both compartments separately. Defaults to True.

        Returns:
            numpy.ndarray: Concentration in M
        """
        # Arterial concentrations
        if self.aif is not None:
            r1 = dc.relaxivity(self.field_strength, 'blood', self.agent)
            if self.sequence == 'SR':
                cb = dc.conc_src(self.aif, self.TC, 1/self.R10b, r1, self.n0)
            elif self.sequence == 'SS':
                cb = dc.conc_ss(self.aif, self.TR, self.FA, 1/self.R10b, r1, self.n0)
            else:
                raise NotImplementedError('Signal model ' + self.sequence + 'is not (yet) supported.') 
            self.ca = cb/(1-self.Hct) 

        t = self.time()

        if self.kinetics == 'non-stationary':
            khe = dc.interp([self.khe*(1-self.Hct), self.khe_f*(1-self.Hct)], t)
            Kbh = dc.interp([1/self.Th, 1/self.Th_f], t)
            return t, dc.conc_liver(self.ca, 
                    self.ve, self.Te, self.De, khe, 1/Kbh,
                    t=self.t, dt=self.dt, kinetics='ICNS', sum=sum)
        
        elif self.kinetics == 'non-stationary uptake':
            self.Th_f = self.Th
            khe = dc.interp([self.khe*(1-self.Hct), self.khe_f*(1-self.Hct)], t)
            return t, dc.conc_liver(self.ca, 
                    self.ve, self.Te, self.De, khe, self.Th,
                    t=self.t, dt=self.dt, kinetics='ICNSU', sum=sum)
        
        elif self.kinetics == 'stationary':
            self.khe_f = self.khe
            self.Th_f = self.Th
            khe = self.khe*(1-self.Hct)
            return t, dc.conc_liver(self.ca, 
                    self.ve, self.Te, self.De, khe, self.Th,
                    t=self.t, dt=self.dt, kinetics='IC', sum=sum)
        
        elif self.kinetics == 'IC-HF':
            self.khe_f = self.khe
            self.Th_f = self.Th
            khe = self.khe*(1-self.Hct)
            return t, dc.conc_liver(self.ca, 
                self.ve, khe, self.Th,
                t=self.t, dt=self.dt, kinetics='IC-HF', sum=sum)

    def relax(self):
        """Tissue relaxation rate

        Returns:
            numpy.ndarray: Relaxation rate in 1/sec
        """
        t, C = self.conc(sum=False)
        rp = dc.relaxivity(self.field_strength, 'blood', self.agent)
        rh = dc.relaxivity(self.field_strength, 'hepatocytes', self.agent)
        return t, self.R10 + rp*C[0,:] + rh*C[1,:]
    
    def predict(self, xdata): 
        t, R1 = self.relax()
        # if self.signal == 'SR':
        #     signal = dc.signal_sr(R1, self.S0, self.TR, self.FA, self.TC, R10=self.R10)
        # else:
        #     signal = dc.signal_ss(R1, self.S0, self.TR, self.FA, R10=self.R10)
        if self.sequence == 'SR':
            signal = dc.signal_sr(R1, self.S0, self.TR, self.FA, self.TC)
        else:
            signal = dc.signal_ss(R1, self.S0, self.TR, self.FA)
        return dc.sample(xdata, t, signal, self.dt)
    
    def train(self, xdata, ydata, **kwargs):
        if self.sequence == 'SR':
            Sref = dc.signal_sr(self.R10, 1, self.TR, self.FA, self.TC)
        else:
            Sref = dc.signal_ss(self.R10, 1, self.TR, self.FA)
        self.S0 = np.mean(ydata[:self.n0]) / Sref
        return dc.train(self, xdata, ydata, **kwargs)
    
    def export_params(self):
        pars = {}
        pars['ve']=["Liver extracellular volume fraction", self.ve, 'mL/mL']
        if self.kinetics=='IC-HF':
            pars['khe']=["Hepatocellular uptake rate", self.khe, 'mL/sec/mL']
            pars['Th']=["Hepatocellular transit time", self.Th, 'sec']
            pars['kbh']=["Biliary excretion rate", (1-self.ve)/self.Th, 'mL/sec/mL']
            pars['Khe']=["Hepatocellular tissue uptake rate", self.khe/self.ve, 'mL/sec/mL']
            pars['Kbh']=["Biliary tissue excretion rate", 1/self.Th, 'mL/sec/mL']
            if self.vol is not None:
                pars['CL']=['Liver blood clearance', self.khe*self.vol, 'mL/sec']
        elif self.kinetics=='stationary':
            pars['Te']=["Extracellular transit time", self.Te, 'sec']
            pars['De']=["Extracellular dispersion", self.De, '']
            pars['khe']=["Hepatocellular uptake rate", self.khe, 'mL/sec/mL']
            pars['Th']=["Hepatocellular transit time", self.Th, 'sec']
            pars['kbh']=["Biliary excretion rate", (1-self.ve)/self.Th, 'mL/sec/mL']
            pars['Khe']=["Hepatocellular tissue uptake rate", self.khe/self.ve, 'mL/sec/mL']
            pars['Kbh']=["Biliary tissue excretion rate", 1/self.Th, 'mL/sec/mL']
            if self.vol is not None:
                pars['CL']=['Liver blood clearance', self.khe*self.vol, 'mL/sec']
        else:
            pars['Te']=["Extracellular transit time", self.Te, 'sec']
            pars['De']=["Extracellular dispersion", self.De, '']
            khe = [self.khe, self.khe_f]
            Kbh = [1/self.Th, 1/self.Th_f]
            khe_avr = np.mean(khe)
            Kbh_avr = np.mean(Kbh)
            khe_var = (np.amax(khe)-np.amin(khe))/khe_avr
            Kbh_var = (np.amax(Kbh)-np.amin(Kbh))/Kbh_avr 
            kbh = np.mean((1-self.ve)*Kbh_avr)
            Th = np.mean(1/Kbh_avr)
            pars['khe']=["Hepatocellular uptake rate", khe_avr, 'mL/sec/mL']
            pars['Th']=["Hepatocellular transit time", Th, 'sec']
            pars['kbh']=["Biliary excretion rate", kbh, 'mL/sec/mL']
            pars['Khe']=["Hepatocellular tissue uptake rate", khe_avr/self.ve, 'mL/sec/mL']
            pars['Kbh']=["Biliary tissue excretion rate", Kbh_avr, 'mL/sec/mL']
            pars['khe_i']=["Hepatocellular uptake rate (initial)", self.khe, 'mL/sec/mL']
            pars['khe_f']=["Hepatocellular uptake rate (final)", self.khe_f, 'mL/sec/mL']
            pars['Th_i']=["Hepatocellular transit time (initial)", self.Th, 'sec']
            pars['Th_f']=["Hepatocellular transit time (final)", self.Th_f, 'sec']
            pars['khe_var']=["Hepatocellular uptake rate variance", khe_var, '']
            pars['Kbh_var']=["Biliary tissue excretion rate variance", Kbh_var, '']
            pars['kbh_i']=["Biliary excretion rate (initial)", (1-self.ve)/self.Th, 'mL/sec/mL']
            pars['kbh_f']=["Biliary excretion rate (final)", (1-self.ve)/self.Th_f, 'mL/sec/mL']
            if self.vol is not None:
                pars['CL']=['Liver blood clearance', khe_avr*self.vol, 'mL/sec']
        return self._add_sdev(pars)

    def plot(self, 
                xdata:tuple[np.ndarray, np.ndarray], 
                ydata:tuple[np.ndarray, np.ndarray],  
                testdata=None, xlim=None, fname=None, show=True):

        time, C = self.conc(sum=True)
        if xlim is None:
            xlim = [np.amin(time), np.amax(time)]
        fig, (ax0, ax1) = plt.subplots(1,2,figsize=(12,5))
        ax0.set_title('Prediction of the MRI signals.')
        ax0.plot(xdata/60, ydata, marker='o', linestyle='None', color='cornflowerblue', label='Data')
        ax0.plot(time/60, self.predict(time), linestyle='-', linewidth=3.0, color='darkblue', label='Prediction')
        ax0.set(xlabel='Time (min)', ylabel='MRI signal (a.u.)', xlim=np.array(xlim)/60)
        ax0.legend()
        ax1.set_title('Reconstruction of concentrations.')
        if testdata is not None:
            ax1.plot(testdata['t']/60, 1000*testdata['C'], marker='o', linestyle='None', color='cornflowerblue', label='Tissue ground truth')
            ax1.plot(testdata['t']/60, 1000*testdata['cp'], marker='o', linestyle='None', color='lightcoral', label='Arterial ground truth')
        ax1.plot(time/60, 1000*C, linestyle='-', linewidth=3.0, color='darkblue', label='Tissue prediction')
        ax1.plot(time/60, 1000*self.ca, linestyle='-', linewidth=3.0, color='darkred', label='Arterial prediction')
        ax1.set(xlabel='Time (min)', ylabel='Concentration (mM)', xlim=np.array(xlim)/60)
        ax1.legend()
        if fname is not None:  
            plt.savefig(fname=fname)
        if show:
            plt.show()
        else:
            plt.close()

    





import matplotlib.pyplot as plt
import numpy as np
import dcmri as dc
from scipy.stats import rv_histogram



class Kidney(dc.Model):
    """Delayed two-compartment filtration model with fast water exchange acquired with a steady-state sequence.

        **Input function**

        - **aif** (array-like, default=None): Signal-time curve in a feeding artery. If AIF is set to None, then the parameter ca must be provided (arterial concentrations).
        - **ca** (array-like, default=None): Concentration (M) in the arterial input. Must be provided when aif = None, ignored otherwise. 

        **Acquisition parameters**

        - **t** (array-like, default=None): Time points (sec) of the aif. If t is not provided, the temporal sampling is uniform with interval dt.
        - **dt** (float, default=1.0): Sampling interval of the AIF in sec.
        - **field_strength** (float, default=3.0): Magnetic field strength in T.
        - **agent** (str, default='gadoterate'): Contrast agent generic name.
        - **n0** (int, default=1): Baseline length in nr of acquisitions.
        - **TR** (float, default=0.005): Repetition time, or time between excitation pulses, in sec.
        - **FA** (float, default=15): Nominal flip angle in degrees.

        **Tracer-kinetic parameters (any model)**

        - **Hct** (float, optional): Hematocrit. 
        - **Fp** (Plasma flow, mL/sec/mL): Flow of plasma into the plasma compartment.
        - **Tp** (Plasma mean transit time, sec): Transit time of the plasma compartment. 
        - **Ft** (Tubular flow, mL/sec/mL): Flow of fluid into the tubuli.

        **Tracer-kinetic parameters (2CFM)**

        - **Ta** (Arterial delay time, sec): Transit time through the arterial compartment.
        - **Tt** (Tubular mean transit time, sec): Transit time of the tubular compartment. 
        
        **Tracer-kinetic parameters (FN)**

        - **h0** (Transit time frequency, in units of 1/sec): Frequency of the transit times in the first bin (default range 15-30s).
        - **h1** (Transit time frequency, in units of 1/sec): Frequency of the transit times in the second bin (default range 30-60s).
        - **h2** (Transit time frequency, in units of 1/sec): Frequency of the transit times in the third bin (default range 60-90s).
        - **h3** (Transit time frequency, in units of 1/sec): Frequency of the transit times in the fourth bin (default range 90-150s).
        - **h4** (Transit time frequency, in units of 1/sec): Frequency of the transit times in the fifth bin (default range 150-300s).
        - **h5** (Transit time frequency, in units of 1/sec): Frequency of the transit times in the sixth bin (default range 300-600s).

        **Signal parameters**

        - **R10** (float, default=1): Precontrast tissue relaxation rate in 1/sec.
        - **R10b** (float, default=1): Precontrast arterial relaxation rate in 1/sec. 
        - **S0** (float, default=1): Scale factor for the MR signal (a.u.).

        **Prediction and training parameters**

        - **free** (array-like): list of free parameters. The default depends on the kinetics parameter.
        - **bounds** (array-like): 2-element list with lower and upper bounds of the free parameters. The default depends on the kinetics parameter.

        **Additional parameters**

        - **vol** (float, optional): Kidney volume in mL.

    See Also:
        `Liver`

    Example:

        Derive model parameters from simulated data:

    .. plot::
        :include-source:
        :context: close-figs
    
        >>> import matplotlib.pyplot as plt
        >>> import dcmri as dc

        Use `fake_tissue` to generate synthetic test data:

        >>> time, aif, roi, gt = dc.fake_tissue(R10=1/dc.T1(3.0,'kidney'))
        
        Override the parameter defaults to match the experimental conditions of the synthetic test data:

        >>> params = {
        ...     'aif':aif, 
        ...     'dt':time[1], 
        ...     'agent': 'gadodiamide', 
        ...     'TR': 0.005, 
        ...     'FA': 20, 
        ...     'R10': 1/dc.T1(3.0,'kidney'), 
        ...     'n0': 15,
        ... }

        Train a two-compartment filtration model on the ROI data and plot the fit:

        >>> params['kinetics'] = '2CFM'
        >>> model = dc.Kidney(**params).train(time, roi)
        >>> model.plot(time, roi, testdata=gt)

        Repeat the fit using a free nephron model:

        >>> params['kinetics'] = 'FN'
        >>> model = dc.Kidney(**params).train(time, roi)
        >>> model.plot(time, roi, testdata=gt)
    """ 

    def __init__(self, **params):

        # Input function
        self.aif = None
        self.ca = None
    
        # Acquisition parameters
        self.sequence = 'SS'
        self.field_strength = 3.0
        self.t = None
        self.dt = 0.5 
        self.agent = 'gadoterate'
        self.n0 = 1
        self.TR = 0.005  
        self.FA = 15.0
        self.Tsat = 0
        self.TC = 0.085
        
        # Tracer-kinetic parameters
        self.kinetics = '2CFM'
        self.Ta = 0
        self.Hct = 0.45
        self.Fp = 200/6000
        self.Tp = 5
        self.Ft = 30/6000

        # 2CFM
        self.Tt = 120
        
        # Free
        self.h0 = 1
        self.h1 = 1
        self.h2 = 1
        self.h3 = 1
        self.h4 = 1
        self.h5 = 1
        self.TT = [15,30,60,90,150,300,600]

        # Signal parameters
        self.R10 = 1/dc.T1(3.0, 'kidney')  
        self.R10b = 1/dc.T1(3.0, 'blood')
        self.S0 = 1

        # Training parameters
        self.free = ['Fp','Tp','Ft','Tt']
        self.bounds = [
            0,
            [np.inf, np.inf, np.inf, np.inf],
        ]
        
        # Training parameters
        if 'kinetics' in params:
            if params['kinetics'] == 'FN':
                self.free = ['Fp','Tp','Ft','h0','h1','h2','h3','h4','h5']
                self.bounds = (
                    0,
                    [np.inf, 8, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf],
                )

        # Additional parameters
        self.vol = None

        # Override defaults
        for k, v in params.items():
            setattr(self, k, v)

        # Check inputs
        if (self.aif is None) and (self.ca is None):
            raise ValueError('Please provide either an arterial sigal (aif) or an arterial concentration (ca).')  


    def time(self):
        """Array of time points.

        Returns:
            np.ndarray: time points in seconds.
        """
        if self.t is None:
            if self.aif is None:
                return self.dt*np.arange(np.size(self.ca))
            else:
                return self.dt*np.arange(np.size(self.aif))
        else:
            return self.t


    def conc(self, sum=True):
        """Tissue concentration

        Args:
            sum (bool, optional): If True, returns the total concentrations. Else returns the concentration in the individual compartments. Defaults to True.

        Returns:
            numpy.ndarray: Concentration in M
        """

        # Arterial concentrations
        if self.aif is not None:
            r1 = dc.relaxivity(self.field_strength, 'blood', self.agent)
            if self.sequence == 'SR':
                cb = dc.conc_src(self.aif, self.TC, 1/self.R10b, r1, self.n0)
            elif self.sequence == 'SS':
                cb = dc.conc_ss(self.aif, self.TR, self.FA, 1/self.R10b, r1, self.n0)
            elif self.sequence == 'lin':
                cb = dc.conc_lin(self.aif, 1/self.R10b, r1, self.n0)
            else:
                raise NotImplementedError('Signal model ' + self.sequence + 'is not (yet) supported.') 
            self.ca = cb/(1-self.Hct) 

        if self.kinetics == '2CFM':
            ca = dc.flux(self.ca, self.Ta, t=self.t, dt=self.dt, kinetics='plug')
            return dc.conc_kidney(ca, self.Fp, self.Tp, self.Ft, self.Tt, t=self.t, dt=self.dt, sum=sum, kinetics='2CF')
        elif self.kinetics == 'FN':
            H = [self.h0,self.h1,self.h2,self.h3,self.h4,self.h5]
            return dc.conc_kidney(self.ca, self.Fp, self.Tp, self.Ft, H, TT=self.TT, t=self.t, dt=self.dt, sum=sum, kinetics='FN')
        
    def relax(self):
        """Longitudinal relaxation rate R1(t).

        Returns:
            np.ndarray: Relaxation rate. Dimensions are (nt,) for a tissue in fast water exchange, or (nc,nt) for a multicompartment tissue outside the fast water exchange limit.
        """
        C = self.conc()
        r1 = dc.relaxivity(self.field_strength, 'blood', self.agent)
        R1 = self.R10 + r1*C
        return R1
        
    def signal(self)->np.ndarray:
        """Pseudocontinuous signal S(t) as a function of time.

        Returns:
            np.ndarray: Signal as a 1D array.
        """
        R1 = self.relax()
        if self.sequence=='SR':
            return dc.signal_sr(R1, self.S0, self.TR, self.FA, self.TC, self.Tsat)
        elif self.sequence == 'SS':
            return dc.signal_ss(R1, self.S0, self.TR, self.FA)
        elif self.sequence == 'lin':
            return dc.signal_lin(R1, self.S0)

    def predict(self, xdata):
        t = self.time()
        if np.amax(xdata) > np.amax(t):
            msg = 'The acquisition window is longer than the duration of the AIF. \n'
            msg += 'The largest time point that can be predicted is ' + str(np.amax(t)/60) + 'min.'
            raise ValueError(msg)
        sig = self.signal()
        return dc.sample(xdata, t, sig, self.dt)

    
    def train(self, xdata, ydata, **kwargs):
        if self.sequence=='SR':
            Sref = dc.signal_sr(self.R10, 1, self.TR, self.FA, self.TC, self.Tsat)
        elif self.sequence == 'SS':
            Sref = dc.signal_ss(self.R10, 1, self.TR, self.FA)
        elif self.sequence == 'lin':
            Sref = dc.signal_lin(self.R10, 1)
        self.S0 = np.mean(ydata[:self.n0]) / Sref
        return dc.train(self, xdata, ydata, **kwargs)

    def export_params(self):
        pars = {}
        pars['S0'] = ['Signal scaling factor', self.S0, 'a.u.']
        pars['Fp'] = ['Plasma flow', self.Fp, 'mL/sec/mL']
        pars['Tp'] = ['Plasma mean transit time', self.Tp, 'sec']
        pars['Ft'] = ['Tubular flow', self.Ft, 'mL/sec/mL']
        pars['Fb'] = ['Blood flow',self.Fp/(1-self.Hct),'mL/sec/mL']
        pars['ve'] = ['Extracellular volume', self.Fp*self.Tp, '']
        pars['FF'] = ['Filtration fraction', self.Ft/self.Fp, '']
        pars['E'] = ['Extraction fraction', self.Ft/(self.Ft+self.Fp), '']
        if self.kinetics == '2CFM':
            pars['Tt'] = ['Tubular mean transit time', self.Tt, 'sec']
            pars['Ta'] = ['Arterial mean transit time', self.Ta, 'sec']
        elif self.kinetics == 'FN':
            H = [self.h0,self.h1,self.h2,self.h3,self.h4,self.h5]
            TT = rv_histogram((H,self.TT), density=True)
            pars['h0'] = ["Transit time weight (15-30s)", self.h0, '1/sec']
            pars['h1'] = ["Transit time weight (30-60s)", self.h1, '1/sec']
            pars['h2'] = ["Transit time weight (60-90s)", self.h2, '1/sec']
            pars['h3'] = ["Transit time weight (90-150s)", self.h3, '1/sec']
            pars['h4'] = ["Transit time weight (150-300s)", self.h4, '1/sec']
            pars['h5'] = ["Transit time weight (300-600s)", self.h5, '1/sec']
            pars['Tt'] = ['Tubular mean transit time', TT.mean(), 'sec']
            pars['Dt'] = ['Tubular transit time dispersion', TT.std()/TT.mean(), '']
        if self.vol is None:
            return self._add_sdev(pars)
        pars['SK-GFR'] = ['Single-kidney glomerular filtration rate', self.Ft*self.vol, 'mL/sec']
        pars['SK-RBF'] = ['Single-kidney renal blood flow', self.Fp*self.vol/(1-self.Hct), 'mL/sec']
        pars['SK-Vol'] = ['Single-kidney volume', self.vol, 'mL']
        return self._add_sdev(pars)


    def plot(self, 
                xdata:tuple[np.ndarray, np.ndarray], 
                ydata:tuple[np.ndarray, np.ndarray],  
                testdata=None, xlim=None, fname=None, show=True):
        time = self.time()
        C = self.conc(sum=True)
        if xlim is None:
            xlim = [np.amin(time), np.amax(time)]
        fig, (ax0, ax1) = plt.subplots(1,2,figsize=(12,5))
        ax0.set_title('Prediction of the MRI signals.')
        ax0.plot(xdata/60, ydata, marker='o', linestyle='None', color='cornflowerblue', label='Data')
        ax0.plot(time/60, self.predict(time), linestyle='-', linewidth=3.0, color='darkblue', label='Prediction')
        ax0.set(xlabel='Time (min)', ylabel='MRI signal (a.u.)', xlim=np.array(xlim)/60)
        ax0.legend()
        ax1.set_title('Reconstruction of concentrations.')
        if testdata is not None:
            ax1.plot(testdata['t']/60, 1000*testdata['C'], marker='o', linestyle='None', color='cornflowerblue', label='Tissue ground truth')
            ax1.plot(testdata['t']/60, 1000*testdata['cp'], marker='o', linestyle='None', color='lightcoral', label='Arterial ground truth')
        ax1.plot(time/60, 1000*C, linestyle='-', linewidth=3.0, color='darkblue', label='Tissue prediction')
        ax1.plot(time/60, 1000*self.ca, linestyle='-', linewidth=3.0, color='darkred', label='Arterial prediction')
        ax1.set(xlabel='Time (min)', ylabel='Concentration (mM)', xlim=np.array(xlim)/60)
        ax1.legend()
        if fname is not None:  
            plt.savefig(fname=fname)
        if show:
            plt.show()
        else:
            plt.close()




class KidneyCortMed(dc.Model):
    """
    Corticomedullary multi-compartment model in fast water exchange acquired with a saturation-recovery sequence.

        **Kinetic parameters**

        - **Hct** (float, optional): Hematocrit. 
        - **Fp** Plasma flow in mL/sec/mL. 
        - **Eg** Glomerular extraction fraction. 
        - **fc** Cortical flow fraction. 
        - **Tg** Glomerular mean transit time in sec. 
        - **Tv** Peritubular & venous mean transit time in sec. 
        - **Tpt** Proximal tubuli mean transit time in sec. 
        - **Tlh** Lis of Henle mean transit time in sec. 
        - **Tdt** Distal tubuli mean transit time in sec. 
        - **Tcd** Collecting duct mean transit time in sec.

        **Input function**

        - **aif** (array-like, default=None): Signal-time curve in a feeding artery. If AIF is set to None, then the parameter ca must be provided (arterial concentrations).
        - **ca** (array-like, default=None): Concentration (M) in the arterial input. Must be provided when aif = None, ignored otherwise.

        **Acquisition parameters**

        - **sequence** (str, default='SR'): imaging sequence.
        - **dt** (float, optional): Sampling interval of the AIF in sec. 
        - **agent** (str, optional): Contrast agent generic name.
        - **field_strength** (float, optional): Magnetic field strength in T. 
        - **TR** (float, optional): Repetition time, or time between excitation pulses, in sec. 
        - **FA** (float, optional): Nominal flip angle in degrees.
        - **Tsat** (float, optional): Time before start of readout (sec).
        - **TC** (float, optional): Time to the center of the readout pulse
        - **n0** (float, optional): Baseline length in number of acquisitions.
       
        **Signal parameters**

        - **R10c** (float, optional): Precontrast cortex relaxation rate in 1/sec. 
        - **R10m** (float, optional): Precontrast medulla relaxation rate in 1/sec.
        - **R10b** (float, optional): Precontrast arterial relaxation rate in 1/sec. 
        - **S0c** (float, optional): Signal scaling factor in the cortex (a.u.).
        - **S0m** (float, optional): Signal scaling factor in the medulla (a.u.).

        **Prediction and training parameters**

        - **free** (array-like): list of free parameters. The default depends on the kinetics parameter.
        - **bounds** (array-like): 2-element list with lower and upper bounds of the free parameters. The default depends on the kinetics parameter.
        
        **Other parameters**

        - **vol** (float, optional): Liver volume in mL.


    See Also:
        `Kidney`, `Liver`

    Example:

        Derive model parameters from simulated data:

    .. plot::
        :include-source:
        :context: close-figs
    
        >>> import dcmri as dc

        Use `fake_kidney_cortex_medulla` to generate synthetic test data:

        >>> time, aif, roi, gt = dc.fake_kidney_cortex_medulla(CNR=100)
        
        Build a tissue model and set the constants to match the experimental conditions of the synthetic test data:

        >>> model = dc.KidneyCortMed(
        ...     aif = aif,
        ...     dt = time[1],
        ...     agent = 'gadoterate',
        ...     TR = 0.005,
        ...     FA = 20,
        ...     TC = 0.2,
        ...     n0 = 10,
        ... )

        Train the model on the ROI data and predict signals and concentrations:

        >>> model.train(time, roi)

        Plot the reconstructed signals (left) and concentrations (right) and compare the concentrations against the noise-free ground truth:

        >>> model.plot(time, roi, testdata=gt)
    """ 



    def __init__(self, **params):

        # Input function
        self.aif = None
        self.ca = None

        # Acquisition parameters
        self.sequence = 'SR'
        self.field_strength = 3.0
        self.t = None
        self.dt = 0.5
        self.agent = 'gadoterate'
        self.n0 = 1
        self.TR = 0.005  
        self.FA = 15.0
        self.Tsat = 0
        self.TC = 0.085

        # Tracer-kinetic parameters
        self.Hct = 0.45 
        self.Fp = 0.03
        self.Eg = 0.15
        self.fc = 0.8
        self.Tg = 4
        self.Tv = 10
        self.Tpt = 60
        self.Tlh = 60
        self.Tdt = 30
        self.Tcd = 30

        # Signal parameters
        self.R10c = 1/dc.T1(3.0, 'kidney')
        self.R10m = 1/dc.T1(3.0, 'kidney')  
        self.R10b = 1/dc.T1(3.0, 'blood')
        self.S0c = 1
        self.S0m = 1
        
        # training parameters
        self.free = ['Fp','Eg','fc','Tg','Tv','Tpt','Tlh','Tdt','Tcd']
        self.bounds = [
            [0.01, 0, 0, 0, 0, 0, 0, 0, 0],
            [1, 1, 1, 10, 30, np.inf, np.inf, np.inf, np.inf],
        ]

        # Other parameters
        self.vol = None

        # Override defaults
        for k, v in params.items():
            setattr(self, k, v)

        # Check inputs
        if (self.aif is None) and (self.ca is None):
            raise ValueError('Please provide either an arterial sigal (aif) or an arterial concentration (ca).')  


    def time(self):
        """Array of time points.

        Returns:
            np.ndarray: time points in seconds.
        """
        if self.t is None:
            if self.aif is None:
                return self.dt*np.arange(np.size(self.ca))
            else:
                return self.dt*np.arange(np.size(self.aif))
        else:
            return self.t
        

    def conc(self, sum=True):
        """Cortical and medullary concentration

        Returns:
            tuple[numpy.ndarray, numpy.ndarray]: Concentration in cortex, concentration in medulla, in M
        """

        if self.aif is not None:
            r1 = dc.relaxivity(self.field_strength, 'blood', self.agent)
            if self.sequence == 'SR':
                cb = dc.conc_src(self.aif, self.TC, 1/self.R10b, r1, self.n0)
            elif self.sequence == 'SS':
                cb = dc.conc_ss(self.aif, self.TR, self.FA, 1/self.R10b, r1, self.n0)
            else:
                raise NotImplementedError('Signal model ' + self.sequence + 'is not (yet) supported.')   
            self.ca = cb/(1-self.Hct)      
        return dc.conc_kidney_cortex_medulla(self.ca, 
            self.Fp, self.Eg, self.fc, self.Tg, self.Tv, 
            self.Tpt, self.Tlh, self.Tdt, self.Tcd, 
            t=self.t, dt=self.dt, sum=sum, kinetics='7C')
        
    def predict(self, xdata:np.ndarray)->tuple[np.ndarray,np.ndarray]:
        Cc, Cm = self.conc()
        r1 = dc.relaxivity(self.field_strength, 'blood', self.agent)
        R1c = self.R10c + r1*Cc
        R1m = self.R10m + r1*Cm
        if self.sequence == 'SR':
            Sc = dc.signal_sr(R1c, self.S0c, self.TR, self.FA, self.TC, self.Tsat)
            Sm = dc.signal_sr(R1m, self.S0m, self.TR, self.FA, self.TC, self.Tsat)
        elif self.sequence == 'SS':
            Sc = dc.signal_ss(R1c, self.S0c, self.TR, self.FA)
            Sm = dc.signal_ss(R1m, self.S0m, self.TR, self.FA)  
        t = self.time()          
        return (
            dc.sample(xdata, t, Sc, self.dt), 
            dc.sample(xdata, t, Sm, self.dt),
        ) 
    
    def train(self, xdata:np.ndarray, ydata:tuple[np.ndarray,np.ndarray], **kwargs):
        if self.sequence == 'SR':
            Scref = dc.signal_sr(self.R10c, 1, self.TR, self.FA, self.TC, self.Tsat)
            Smref = dc.signal_sr(self.R10m, 1, self.TR, self.FA, self.TC, self.Tsat)
        elif self.sequence == 'SS':
            Scref = dc.signal_ss(self.R10c, 1, self.TR, self.FA)
            Smref = dc.signal_ss(self.R10m, 1, self.TR, self.FA)        
        self.S0c = np.mean(ydata[0][:self.n0]) / Scref
        self.S0m = np.mean(ydata[1][:self.n0]) / Smref
        return dc.train(self, xdata, ydata, **kwargs)


    def export_params(self):
        pars = {}
        pars['Fp']=['Plasma flow',self.Fp,'mL/sec/mL']
        pars['Eg']=['Glomerular extraction fraction',self.Eg,'']
        pars['fc']=['Cortical flow fraction',self.fc,'']
        pars['Tg']=['Glomerular mean transit time',self.Tg,'sec']
        pars['Tv']=['Peritubular & venous mean transit time',self.Tv,'sec']
        pars['Tpt']=['Proximal tubuli mean transit time',self.Tpt,'sec'] 
        pars['Tlh']=['Lis of Henle mean transit time',self.Tlh,'sec'] 
        pars['Tdt']=['Distal tubuli mean transit time',self.Tdt,'sec'] 
        pars['Tcd']=['Collecting duct mean transit time',self.Tcd,'sec']
        pars['FF']=['Filtration fraction', self.Eg/(1-self.Eg), '']
        pars['Ft']=['Tubular flow', self.Fp*self.Eg/(1-self.Eg), 'mL/sec/mL']
        pars['CBF']=['Cortical blood flow', self.Fp/(1-self.Hct), 'mL/sec/mL']
        pars['MBF']=['Medullary blood flow', (1-self.fc)*(1-self.Eg)*self.Fp/(1-self.Hct), 'mL/sec/mL']
        if self.vol is None:   
            return self._add_sdev(pars)
        pars['SKGFR']=['Single-kidney glomerular filtration rate', self.vol*self.Fp*self.Eg/(1-self.Eg),'mL/sec']
        pars['SKBF']=['Single-kidney blood flow', self.vol*self.Fp/(1-self.Hct), 'mL/sec']
        pars['SKMBF']=['Single-kidney medullary blood flow', self.vol*(1-self.fc)*(1-self.Eg)*self.Fp/(1-self.Hct), 'mL/sec']

        return self._add_sdev(pars)


    def plot(self, xdata:np.ndarray, ydata:tuple[np.ndarray,np.ndarray], 
             testdata=None, xlim=None, fname=None, show=True):

        time = self.time()
        if xlim is None:
            xlim = [np.amin(time), np.amax(time)]
        roic_pred, roim_pred = self.predict(time)
        concc, concm = self.conc()

        fig, (ax0, ax1) = plt.subplots(1,2,figsize=(12,5))
        ax0.set_title('Prediction of the MRI signals.')
        ax0.plot(xdata/60, ydata[0], marker='o', linestyle='None', color='cornflowerblue', label='Cortex data')
        ax0.plot(xdata/60, ydata[1], marker='x', linestyle='None', color='cornflowerblue', label='Medulla data')
        ax0.plot(time/60, roic_pred, linestyle='-', linewidth=3.0, color='darkblue', label='Cortex prediction')
        ax0.plot(time/60, roim_pred, linestyle='--', linewidth=3.0, color='darkblue', label='Medulla prediction')
        ax0.set(xlabel='Time (min)', ylabel='MRI signal (a.u.)', xlim=np.array(xlim)/60)
        ax0.legend()

        ax1.set_title('Reconstruction of concentrations.')
        if testdata is not None:
            ax1.plot(testdata['t']/60, 1000*testdata['Cc'], marker='o', linestyle='None', color='cornflowerblue', label='Cortex ground truth')
            ax1.plot(testdata['t']/60, 1000*testdata['Cm'], marker='x', linestyle='None', color='cornflowerblue', label='Medulla ground truth')
            ax1.plot(testdata['t']/60, 1000*testdata['cp'], marker='o', linestyle='None', color='lightcoral', label='Arterial ground truth')
        ax1.plot(time/60, 1000*concc, linestyle='-', linewidth=3.0, color='darkblue', label='Cortex prediction')
        ax1.plot(time/60, 1000*concm, linestyle='--', linewidth=3.0, color='darkblue', label='Medulla prediction')
        ax1.plot(time/60, 1000*self.ca, linestyle='-', linewidth=3.0, color='darkred', label='Arterial prediction')
        ax1.set(xlabel='Time (min)', ylabel='Concentration (mM)', xlim=np.array(xlim)/60)
        ax1.legend()

        if fname is not None:  
            plt.savefig(fname=fname)
        if show:
            plt.show()
        else:
            plt.close()



import matplotlib.pyplot as plt
import numpy as np
import dcmri as dc


class Tissue(dc.Model):
    """Model for general vascular-interstitial tissues with bi-directional exchange of indicator.

        **Input function**

        - **aif** (array-like, default=None): Signal-time curve in a feeding artery. If AIF is set to None, then the parameter ca must be provided (arterial concentrations).
        - **ca** (array-like, default=None): Concentration (M) in the arterial input. Must be provided when aif = None, ignored otherwise. 

        **Acquisition parameters**

        - **t** (array-like, default=None): Time points (sec) of the aif. If t is not provided, the temporal sampling is uniform with interval dt.
        - **dt** (float, default=1.0): Sampling interval of the AIF in sec.
        - **field_strength** (float, default=3.0): Magnetic field strength in T.
        - **agent** (str, default='gadoterate'): Contrast agent generic name.
        - **n0** (int, default=1): Baseline length in nr of acquisitions.
        - **TR** (float, default=0.005): Repetition time, or time between excitation pulses, in sec.
        - **FA** (float, default=15): Nominal flip angle in degrees.
        - **TC** (float, default=0.1): Time to the center of k-space in a saturation-recovery sequence.
        - **TP** (float, default=0): Preparation delay in a saturation-recovery sequence.

        **Tracer-kinetic parameters**

        - **kinetics** (str, default='HF'): Tracer-kinetic model (see below for options)
        - **Hct** (float, default=0.45): Hematocrit.
        - **Fp** (float, default=0.01): Plasma flow, or flow of plasma into the plasma compartment (mL/sec/mL).
        - **PS** (float, default=0.003): Permeability-surface area product: volume of plasma cleared of indicator per unit time and per unit tissue (mL/sec/mL).
        - **vp** (float, default=0.1): Plasma volume, or volume fraction of the plasma compartment (mL/mL). 
        - **ve** (float, default=0.5): Extravascular, extracellular volume: volume fraction of the interstitial compartment (mL/mL).
        - **Ktrans** (float, default=0.0023): Volume transfer constant (mL/sec/mL).
        - **v** (float, default=0.6): Extracellular volume fraction (mL/mL).

        **Water-kinetic parameters**

        - **water_exchange** (str, default='fast'): Water exchange regime ('fast', 'none' or 'any').
        - **PSe** (float, default=10): Transendothelial water permeability-surface area product: PS for water across the endothelium (mL/sec/mL).
        - **PSc** (float, default=10): Transcytolemmal water permeability-surface area product: PS for water across the cell wall (mL/sec/mL).

        **Signal parameters**

        - **sequence** (str, default='SS'): imaging sequence.
        - **R10** (float, default=1): Precontrast tissue relaxation rate in 1/sec.
        - **R10b** (float, default=1): Precontrast arterial relaxation rate in 1/sec. 
        - **S0** (float, default=1): Scale factor for the MR signal (a.u.).

        **Prediction and training parameters**

        - **free** (array-like): list of free parameters. The default depends on the kinetics parameter.
        - **bounds** (array-like): 2-element list with lower and upper bounds of the free parameters. The default depends on the kinetics parameter.

    Args:
        params (dict, optional): override defaults for any of the parameters.

    Notes:

        Possible values for the **kinetics** argument, along with relevant parameters:: 
        
        - 'U': uptake tissue. Parameters: Fp
        - 'NX': no tracer exchange tissue. Parameters: Fp, vp
        - 'FX': fast tracer exchange tissue. Parameters: Fp, v
        - 'WV': weakly vascularized tissue - also known as *Tofts model*.Parameters: Ktrans, ve
        - 'HFU': high-flow uptake tissue - also known as *Patlak model*. Parameters: vp, PS
        - 'HF': high-flow tissue - also known as *extended Tofts model*, *extended Patlak model* or *general kinetic model*. Params = (vp, PS, ve, )
        - '2CUM': two-compartment uptake tissue. Parameters: Fp, vp, PS
        - '2CXM': two-compartment exchange tissue. Parameters: Fp, vp, PS, ve

    See Also:
        `Liver`, `Kidney`

    Example:

        Fit extended Tofts model to data.

    .. plot::
        :include-source:
        :context: close-figs
    
        >>> import matplotlib.pyplot as plt
        >>> import dcmri as dc

        Use `fake_tissue` to generate synthetic test data:

        >>> time, aif, roi, gt = dc.fake_tissue(CNR=50)
        
        Build a tissue model and set the constants to match the experimental conditions of the synthetic test data:

        >>> model = dc.Tissue(
        ...     aif = aif,
        ...     dt = time[1],
        ...     agent = 'gadodiamide',
        ...     TR = 0.005,
        ...     FA = 20,
        ...     n0 = 15,
        ... )

        Train the model on the ROI data:

        >>> model.train(time, roi)

        Plot the reconstructed signals (left) and concentrations (right) and compare the concentrations against the noise-free ground truth:

        >>> model.plot(time, roi, testdata=gt)
    """ 

    def __init__(self, **params):

        # Input function
        self.aif = None
        self.ca = None
        
        # Acquisition parameters
        self.sequence = 'SS'
        self.field_strength = 3.0
        self.t = None
        self.dt = 1.0
        self.agent = 'gadoterate'
        self.n0 = 1
        self.TR = 0.005
        self.FA = 15
        self.TC = 0.1
        self.TP = 0
        
        # Tracer-kinetic parameters
        self.kinetics = 'HF'
        self.Hct = 0.45
        self.Fp = 0.01
        self.PS = 0.003
        self.vp = 0.1
        self.ve = 0.5

        # Water-kinetic parameters (TODO: what are typical values?)
        self.water_exchange = 'fast'
        self.PSe = 10
        self.PSc = 10

        # Signal parameters
        self.R10 = 1/dc.T1(3.0, 'muscle')
        self.R10b = 1/dc.T1(3.0, 'blood')
        self.S0 = 1

        # training parameters
        self.free = ['PS','vp','ve']
        self.bounds = [0, [np.inf, 1, 1]]

        # Preset parameters
        if 'kinetics' in params:
            if params['kinetics'] == '2CXM': # TODO: change to 2CX
                self.free = ['Fp','PS','vp','ve']
                self.bounds = [0, [np.inf, np.inf, 1, 1]]
            elif params['kinetics'] == '2CUM': # TODO: change to 2CU
                self.free = ['Fp','PS','vp']
                self.bounds = [0, [np.inf, np.inf, 1]]
            elif params['kinetics'] == 'HF':
                self.free = ['PS','vp','ve']
                self.bounds = [0, [np.inf, 1, 1]]
            elif params['kinetics'] == 'HFU':
                self.free = ['Fp','vp']
                self.bounds = [0, [np.inf, 1]]
            elif params['kinetics'] == 'WV': 
                self.free = ['Ktrans','ve']
                self.bounds = [0, [np.inf, 1]]
            elif params['kinetics'] == 'FX':
                self.free = ['Fp','v']
                self.bounds = [0, [np.inf, 1]] 
            elif params['kinetics'] == 'NX': 
                self.free = ['Fp','vp']
                self.bounds = [0, [np.inf, 1]]
            elif params['kinetics'] == 'U': 
                self.free = ['Fp']
                self.bounds = [0, np.inf]

        if 'water_exchange' in params:        
            if params['water_exchange'] == 'any':
                self.free += ['PSe', 'PSc']
                self.bounds[1] += [np.inf, np.inf]

        # Override defaults
        for k, v in params.items():
            setattr(self, k, v)

        # Check inputs
        if (self.aif is None) and (self.ca is None):
            raise ValueError('Please provide either an arterial sigal (aif) or an arterial concentration (ca).')  
        
        # Dependent parameters
        if 'Ktrans' not in params.items():
            self.Ktrans = self.PS*self.Fp/(self.PS+self.Fp)
        if 'v' not in params.items():
            self.v = self.vp+self.ve


    def time(self):
        """Array of time points.

        Returns:
            np.ndarray: time points in seconds.
        """
        if self.t is None:
            if self.aif is None:
                return self.dt*np.arange(np.size(self.ca))
            else:
                return self.dt*np.arange(np.size(self.aif))
        else:
            return self.t


    def conc(self, sum=True):
        """Tissue concentration

        Args:
            sum (bool, optional): If True, returns the total concentrations. Else returns the concentration in the individual compartments. Defaults to True.

        Returns:
            np.ndarray: Concentration in M
        """

        # Arterial concentrations
        if self.aif is not None:
            r1 = dc.relaxivity(self.field_strength, 'blood', self.agent)
            if self.sequence == 'SR':
                cb = dc.conc_src(self.aif, self.TC, 1/self.R10b, r1, self.n0)
            elif self.sequence == 'SS':
                cb = dc.conc_ss(self.aif, self.TR, self.FA, 1/self.R10b, r1, self.n0)
            else:
                raise NotImplementedError('Signal model ' + self.sequence + 'is not (yet) supported.') 
            self.ca = cb/(1-self.Hct) 

        # Tissue concentrations
        if self.kinetics == '2CXM': # (Fp, PS, vp, ve)
            self.Ktrans = self.Fp*self.PS/(self.Fp+self.PS)
            self.v = self.vp + self.ve
            C = dc.conc_tissue(self.ca, self.Fp, self.vp, self.PS, self.ve, t=self.t, dt=self.dt, sum=sum, kinetics='2CX')
            #C = dc._conc_2cx(self.ca, self.Fp, self.vp, self.PS, self.ve, self.t, dt=self.dt, sum=sum)
        elif self.kinetics == '2CUM': # (Fp, PS, vp)
            self.Ktrans = self.Fp*self.PS/(self.Fp+self.PS)
            C = dc.conc_tissue(self.ca, self.Fp, self.vp, self.PS, t=self.t, dt=self.dt, sum=sum, kinetics='2CU')
            #C = dc._conc_2cu(self.ca, self.Fp, self.vp, self.PS, self.t, dt=self.dt, sum=sum)
        elif self.kinetics == 'HF': # (PS, vp, ve)
            self.Fp = np.inf
            self.Ktrans = self.PS
            self.v = self.ve + self.vp
            C = dc.conc_tissue(self.ca, self.vp, self.PS, self.ve, t=self.t, dt=self.dt, sum=sum, kinetics='HF')
            #C = dc.conc_hf(self.ca, self.vp, self.PS, self.ve, self.t, dt=self.dt, sum=sum)
        elif self.kinetics == 'HFU': # (Fp, PS, vp)
            self.Fp = np.inf
            self.Ktrans = self.PS
            C = dc.conc_tissue(self.ca, self.vp, self.PS, t=self.t, dt=self.dt, sum=sum, kinetics='HFU')
            # C = dc._conc_hfu(self.ca, self.vp, self.PS, self.t, dt=self.dt, sum=sum)  
        elif self.kinetics == 'WV': # (Ktrans, ve)
            self.v = self.ve
            self.vp = 0    
            C = dc.conc_tissue(self.ca, self.Ktrans, self.ve, t=self.t, dt=self.dt, sum=sum, kinetics='FX')        
            #C = dc._conc_1c(self.ca, self.Ktrans, self.ve, self.t, dt=self.dt)
        elif self.kinetics == 'FX': # (Fp, v)
            self.PS = np.inf
            self.Ktrans = self.Fp
            C = dc.conc_tissue(self.ca, self.Fp, self.v, t=self.t, dt=self.dt, kinetics='FX')
            #C = dc._conc_1c(self.ca, self.Fp, self.v, self.t, dt=self.dt)
        elif self.kinetics == 'NX': # (Fp, vp)
            C = dc.conc_tissue(self.ca, self.Fp, self.vp, t=self.t, dt=self.dt, kinetics='NX')
            #C = dc._conc_1c(self.ca, self.Fp, self.vp, self.t, dt=self.dt)
        elif self.kinetics == 'U': # (Fp)
            C = dc.conc_tissue(self.ca, self.Fp, t=self.t, dt=self.dt, kinetics='U')
            #C = dc._conc_u(self.ca, self.Fp, self.t, dt=self.dt)
        return C


    def relax(self):
        """Longitudinal relaxation rate R1(t).

        Returns:
            np.ndarray: Relaxation rate. Dimensions are (nt,) for a tissue in fast water exchange, or (nc,nt) for a multicompartment tissue outside the fast water exchange limit.
        """
        C = self.conc(sum=False)
        r1 = dc.relaxivity(self.field_strength, 'blood', self.agent)
        if self.water_exchange == 'fast':
            if self.kinetics in ['FX', 'NX', 'U']:
                R1 = self.R10 + r1*C
            else:
                R1 = self.R10 + r1*(C[0,:]+C[1,:])
        elif self.kinetics in ['2CXM', 'HF']:
            vb = self.vp/(1-self.Hct)
            R1b = self.R10b + r1*C[0,:]/vb
            R1e = self.R10 + r1*C[1,:]/self.ve
            R1c = self.R10 + np.zeros(C.shape[1])
            R1 = np.stack((R1b, R1e, R1c))
        elif self.kinetics in ['2CUM','HFU']:
            vb = self.vp/(1-self.Hct)
            R1b = self.R10b + r1*C[0,:]/vb
            R1e = self.R10 + r1*C[1,:]/(1-vb)
            R1 = np.stack((R1b, R1e))    
        elif self.kinetics in ['WV','FX']:
            R1e = self.R10 + r1*C/self.v
            R1c = self.R10 + np.zeros(C.size)
            R1 = np.stack((R1e, R1c))
        elif self.kinetics == 'NX':
            vb = self.vp/(1-self.Hct)
            R1b = self.R10b + r1*C/vb
            R1e = self.R10 + np.zeros(C.size)
            R1 = np.stack((R1b, R1e))  
        elif self.kinetics == 'U':
            R1 = self.R10 + r1*C
        return R1


    def signal(self)->np.ndarray:
        """Pseudocontinuous signal S(t) as a function of time.

        Returns:
            np.ndarray: Signal as a 1D array.
        """
        R1 = self.relax()

        if self.water_exchange == 'fast':
            if self.sequence == 'SR':
                return dc.signal_sr(R1, self.S0, self.TR, self.FA, self.TC, self.TP)
            elif self.sequence == 'SS':
                return dc.signal_ss(R1, self.S0, self.TR, self.FA)

        elif self.water_exchange == 'none':

            if self.kinetics in ['2CXM','HF']:
                vb = self.vp/(1-self.Hct)
                v = [vb, self.ve, 1-vb-self.ve]
                if self.sequence == 'SR':
                    return dc.signal_sr_nex(v, R1, self.S0, self.TR, self.FA, self.TC, self.TP)
                elif self.sequence == 'SS':
                    return dc.signal_ss(R1, self.S0, self.TR, self.FA, v=v, PSw=0)

            elif self.kinetics in ['2CUM','HFU','NX']:
                vb = self.vp/(1-self.Hct)
                v = [vb, 1-vb]
                if self.sequence == 'SR':
                    return dc.signal_sr_nex(v, R1, self.S0, self.TR, self.FA, self.TC, self.TP)
                elif self.sequence == 'SS':
                    return dc.signal_ss(R1, self.S0, self.TR, self.FA, v=v, PSw=0)

            elif self.kinetics in ['WV','FX']:
                v = [self.ve, 1-self.ve]
                if self.sequence == 'SR':
                    return dc.signal_sr_nex(v, R1, self.S0, self.TR, self.FA, self.TC, self.TP)
                elif self.sequence == 'SS':
                    return dc.signal_ss(R1, self.S0, self.TR, self.FA, v=v, PSw=0)

            elif self.kinetics == 'U':
                if self.sequence == 'SR':
                    return dc.signal_sr(R1, self.S0, self.TR, self.FA, self.TC, self.TP)
                elif self.sequence == 'SS':
                    return dc.signal_ss(R1, self.S0, self.TR, self.FA)

        elif self.water_exchange == 'any':

            if self.kinetics in ['2CXM','HF']:
                vb = self.vp/(1-self.Hct)
                v = [vb, self.ve, 1-vb-self.ve]
                PSw = np.array([[0,self.PSe,0],[self.PSe,0,self.PSc],[0,self.PSc,0]])
                # TODO: ADD diagonal element (flow term)!!
                # Also needs adding inflow 
                # Fb = self.Fp/(1-self.Hct)
                # PSw = np.array([[Fb,self.PSe,0],[self.PSe,0,self.PSc],[0,self.PSc,0]])
                if self.sequence == 'SR':
                    return NotImplementedError('Saturation-recovery not yet implemented for intermediate water exchange')
                elif self.sequence == 'SS':
                    return dc.signal_ss(R1, self.S0, self.TR, self.FA, v=v, PSw=PSw)

            elif self.kinetics in ['2CUM','HFU','NX']:
                vb = self.vp/(1-self.Hct)
                v = [vb, 1-vb]
                # TODO: ADD diagonal element (flow term)!!
                # Fb = self.Fp/(1-self.Hct)
                # PSw = np.array([[Fb,self.PSe],[self.PSe,0]])
                PSw = np.array([[0,self.PSe],[self.PSe,0]])
                if self.sequence == 'SR':
                    return NotImplementedError('Saturation-recovery not yet implemented for intermediate water exchange')
                elif self.sequence == 'SS':
                    return dc.signal_ss(R1, self.S0, self.TR, self.FA, v=v, PSw=PSw)

            elif self.kinetics in ['WV','FX']:
                v = [self.ve, 1-self.ve]
                PSw = np.array([[0,self.PSc],[self.PSc,0]])
                # TODO: ADD diagonal element (flow term)!!
                # FX:
                # Fb = self.Fp/(1-self.Hct)
                # PSw = np.array([[Fb,self.PSc],[self.PSc,0]])
                # WV:
                # PSw = np.array([[self.PSe,self.PSc],[self.PSc,0]])
                if self.sequence == 'SR':
                    return NotImplementedError('Saturation-recovery not yet implemented for intermediate water exchange')
                elif self.sequence == 'SS':
                    return dc.signal_ss(R1, self.S0, self.TR, self.FA, v=v, PSw=PSw)

            elif self.kinetics == 'U':
                if self.sequence == 'SR':
                    return dc.signal_sr(R1, self.S0, self.TR, self.FA, self.TC, self.TP)
                elif self.sequence == 'SS':
                    return dc.signal_ss(R1, self.S0, self.TR, self.FA)
                

    def predict(self, xdata:np.ndarray)->np.ndarray:
        t = self.time()
        if np.amax(xdata) > np.amax(t):
            msg = 'The acquisition window is longer than the duration of the AIF. \n'
            msg += 'The largest time point that can be predicted is ' + str(np.amax(t)/60) + 'min.'
            raise ValueError(msg)
        sig = self.signal()
        return dc.sample(xdata, t, sig, self.dt)
   
    def train(self, xdata, ydata, **kwargs):
        if self.sequence == 'SR':
            Sref = dc.signal_sr(self.R10, 1, self.TR, self.FA, self.TC, self.TP)
        elif self.sequence == 'SS':
            Sref = dc.signal_ss(self.R10, 1, self.TR, self.FA)
        else:
            raise NotImplementedError('Signal model ' + self.sequence + 'is not (yet) supported.') 
        self.S0 = np.mean(ydata[:self.n0]) / Sref
        return dc.train(self, xdata, ydata, **kwargs)
    
    def export_params(self):
        vb = self.vp/(1-self.Hct)
        pars = {
            'Fp': ['Plasma flow',self.Fp,'mL/sec/mL'],
            'PS': ['Permeability-surface area product',self.PS,'mL/sec/mL'],
            'Ktrans': ['Volume transfer constant',self.Ktrans,'mL/sec/mL'],
            've': ['Extravascular extracellular volume',self.ve,'mL/mL'],
            'vb': ['Blood volume',vb,'mL/mL'],
            'vp': ['Plasma volume',self.vp,'mL/mL'],
            'v': ['Extracellular volume',self.v,'mL/mL'],
            'Te': ['Extracellular mean transit time',self.ve/self.PS,'sec'],
            'kep': ['Extravascular transfer constant',self.Ktrans/self.ve,'1/sec'],
            'E': ['Extraction fraction',self.PS/(self.PS+self.Fp),''],
            'Tp': ['Plasma mean transit time',self.vp/(self.Fp+self.PS),'sec'],
            'Tb': ['Blood mean transit time',self.vp/self.Fp,'sec'],
            'T': ['Mean transit time',self.v/self.Fp,'sec'],
            'PSe': ['Transendothelial water PS',self.PSe,'mL/sec/mL'],
            'PSc': ['Transcytolemmal water PS',self.PSc,'mL/sec/mL'],
            'Twc': ['Intracellular water mean transit time', (1-vb-self.ve)/self.PSc, 'sec'],
            'Twi': ['Interstitial water mean transit time', self.ve/(self.PSc+self.PSe), 'sec'],
            'Twb': [ 'Intravascular water mean transit time', vb/self.PSe, 'sec'],
        }
        return self._add_sdev(pars)
    
    def plot(self, 
                xdata:tuple[np.ndarray, np.ndarray], 
                ydata:tuple[np.ndarray, np.ndarray],  
                testdata=None, xlim=None, fname=None, show=True):
        time = self.time()
        C = self.conc(sum=True)
        if xlim is None:
            xlim = [np.amin(time), np.amax(time)]
        fig, (ax0, ax1) = plt.subplots(1,2,figsize=(12,5))
        ax0.set_title('Prediction of the MRI signals.')
        ax0.plot(xdata/60, ydata, marker='o', linestyle='None', color='cornflowerblue', label='Data')
        ax0.plot(time/60, self.predict(time), linestyle='-', linewidth=3.0, color='darkblue', label='Prediction')
        ax0.set(xlabel='Time (min)', ylabel='MRI signal (a.u.)', xlim=np.array(xlim)/60)
        ax0.legend()
        ax1.set_title('Reconstruction of concentrations.')
        if testdata is not None:
            ax1.plot(testdata['t']/60, 1000*testdata['C'], marker='o', linestyle='None', color='cornflowerblue', label='Tissue ground truth')
            ax1.plot(testdata['t']/60, 1000*testdata['cp'], marker='o', linestyle='None', color='lightcoral', label='Arterial ground truth')
        ax1.plot(time/60, 1000*C, linestyle='-', linewidth=3.0, color='darkblue', label='Tissue prediction')
        ax1.plot(time/60, 1000*self.ca, linestyle='-', linewidth=3.0, color='darkred', label='Arterial prediction')
        ax1.set(xlabel='Time (min)', ylabel='Concentration (mM)', xlim=np.array(xlim)/60)
        ax1.legend()
        if fname is not None:
            plt.savefig(fname=fname)
        if show:
            plt.show()
        else:
            plt.close()




import math
from scipy.linalg import expm
import numpy as np


def signal_dsc(R1, R2, S0:float, TR, TE)->np.ndarray:
    """Signal model for a DSC scan with T2 and T2-weighting.

    Args:
        R1 (array-like): Longitudinal relaxation rate in 1/sec. Must have the same size as R2.
        R2 (array-like): Transverse relaxation rate in 1/sec. Must have the same size as R1.
        S0 (float): Signal scaling factor (arbitrary units).
        TR (array-like): Repetition time, or time between successive selective excitations, in sec. If TR is an array, it must have the same size as R1 and R2.
        TE (array-like): Echo time, in sec. If TE is an array, it must have the same size as R1 and R2.

    Returns:
        np.ndarray: Signal in arbitrary units, same length as R1 and R2.
    """
    return S0*np.exp(-np.multiply(TE,R2))*(1-np.exp(-np.multiply(TR,R1)))


def signal_t2w(R2, S0:float, TE)->np.ndarray:
    """Signal model for a DSC scan with T2-weighting.

    Args:
        R2 (array-like): Transverse relaxation rate in 1/sec. Must have the same size as R1.
        S0 (float): Signal scaling factor (arbitrary units).
        TE (array-like): Echo time, in sec. If TE is an array, it must have the same size as R1 and R2.

    Returns:
        np.ndarray: Signal in arbitrary units, same length as R1 and R2.
    """
    return S0*np.exp(-np.multiply(TE,R2))

def conc_t2w(S, TE:float, r2=0.5, n0=1)->np.ndarray:
    """Concentration for a DSC scan with T2-weighting.

    Args:
        S (array-like): Signal in arbitrary units.
        TE (float): Echo time in sec.
        r2 (float, optional): Transverse relaxivity in Hz/M. Defaults to 0.5.
        n0 (int, optional): Baseline length. Defaults to 1.

    Returns:
        np.ndarray: Concentration in M, same length as S.
    """
    # S/Sb = exp(-TE(R2-R2b))
    #   ln(S/Sb) = -TE(R2-R2b)
    #   R2-R2b = -ln(S/Sb)/TE
    # R2 = R2b + r2C
    #   C = (R2-R2b)/r2
    #   C = -ln(S/Sb)/TE/r2
    Sb = np.mean(S[:n0])
    C = -np.log(S/Sb)/TE/r2
    return C



def _signal_ss(R1, Sinf, TR, FA)->np.ndarray:
    E = np.exp(-TR*R1)
    cFA = np.cos(FA*np.pi/180)
    return Sinf * (1-E) / (1-cFA*E)

def _signal_ss_fex(v, R1, S0:float, TR:float, FA:float):
    if np.size(R1) == np.size(v):
        R1 = np.sum(np.multiply(v,R1))
        return _signal_ss(R1, S0, TR, FA)
    nc = R1.shape[0]
    R1fex = np.zeros(R1.shape[1:])
    for c in range(nc):
        R1fex += v[c]*R1[c,...]
    return _signal_ss(R1fex, S0, TR, FA)

def _signal_ss_nex(v, R1:np.ndarray, S0:float, TR:float, FA:float):
    if np.size(R1) == np.size(v):
        S = signal_ss(R1, S0, TR, FA)
        return np.sum(np.multiply(v,S)) 
    nc, nt = R1.shape
    S = np.zeros(nt)
    for c in range(nc):
        S += v[c]*signal_ss(R1[c,:], S0, TR, FA)
    return S

def _signal_ss_aex(PS, v, R1, S0, TR, FA):

    # Mathematical notes on water exchange modelling
    # ---------------------------------------------
    # Longitudinal magnetization in a 2-compartment system
    # with kinetic transport and free relaxation

    # v1*dm1/dt = fi1*mi1 - fo1*m1 + f12*m2 - f21*m1 + R1_1*v1*(m01-m1) 
    # v2*dm2/dt = fi2*mi2 - fo2*m2 + f21*m1 - f12*m2 + R1_2*v2*(m02-m2) 

    # v1*dm1/dt = fi1*mi1 - (fo1+f21+R1_1*v1)*m1 + f12*m2 + R1_1*v1*m01
    # v2*dm2/dt = fi2*mi2 - (fo2+f12+R1_2*v2)*m2 + f21*m1 + R1_2*v2*m02

    # f1 = fo1 + f21 + R1_1*v1 > 0
    # f2 = fo2 + f12 + R1_2*v2 > 0

    # J1(t) = fi1*mi1(t) + R1_1*v1*m01 > 0
    # J2(t) = fi2*mi2(t) + R1_2*v2*m02 > 0

    # v1*dm1/dt = J1 - f1*m1 + f12*m2 
    # v2*dm2/dt = J2 - f2*m2 + f21*m1 

    # K1 = (fo1 + f21)/v1 + R1_1
    # K2 = (fo2 + f12)/v2 + R1_2
    # K12 = f12/v2
    # K21 = f21/v1

    # dM1/dt = J1 - K1*M1 + K12*M2 
    # dM2/dt = J2 - K2*M2 + K21*M1 

    # K = [[K1, -K12],[-K21, K2]
    # dM/dt = J - KM

    # Use generic solution for n-comp system to solve for M(t)
    # Note with R1(t) this is not a stationary system

    # Solution with K and J constant in time:

    # M(t) = exp(-tK)M(0) + exp(-tK)*J
    # M(t) = exp(-tK)M(0) + (1-exp(-tK)) K^-1 J

    # Check

    # dM/dt 
    # = - K exp(-tK)M(0) + exp(-tK) J
    # J - KM 
    # = J - Kexp(-tK)M(0) - (1-exp(-tK))J 
    # = - K exp(-tK)M(0) + exp(-tK) J

    # Spoiled gradient echo steady state:
    # M = exp(-TR*K) cosFA M + (1-exp(-TR*K)) K^-1 J
    # (1 - cosFA exp(-TR*K)) M = (1-exp(-TR*K)) K^-1 J
    # M = (1 - cosFA exp(-TR*K))^-1 (1-exp(-TR*K)) K^-1 J
    # M = (1 - cosFA exp(-TR*K))^-1 K^-1 (1-exp(-TR*K))  J
    # M = [K(1 - cosFA exp(-TR*K))]^-1 (1-exp(-TR*K)) J

    # Inputs to the function:

    # f1 = fo1 + f21 + R1_1(t)*v1
    # f2 = fo2 + f12 + R1_2(t)*v2
    
    # J1 = fi1*S0*mi1(t) + R1_1(t)*v1*S0*m01
    # J2 = fi2*S0*mi2(t) + R1_2(t)*v2*S0*m02

    # K = [[f1/v1, -f21/v1], [-f12/v2, f2/v2]]
    # J = [J1, J2]

    # K dimensions = (ncomp, ncomp, nt)
    # J dimensions = (ncomp, nt)

    # Returns:

    # M = (1 - cosFA exp(-TR*K))^-1 (1-exp(-TR*K)) K^-1 J
    # M = [(1 - cosFA exp(-TR*K))K]^-1 (1-exp(-TR*K)) J

    # reshape for convenience
    n = np.shape(R1)
    R1 = np.reshape(R1, (n[0],-1))
    PS = np.array(PS)

    nc, nt = R1.shape
    J = np.empty((nc,nt))
    K = np.empty((nc,nc,nt))
    for c in range(nc):
        J[c,:] = S0*v[c]*R1[c,:]
        K[c,c,:] = R1[c,:] + np.sum(PS[:,c])/v[c]
        for d in range(nc):
            if d!=c:
                K[c,d,:] = -PS[c,d]/v[d]

    cFA = np.cos(FA*np.pi/180)
    Mag = np.empty(nt)
    Id = np.eye(nc)
    for t in range(nt):
        Et = expm(-TR*K[:,:,t])
        Mt = np.dot(K[:,:,t], Id-cFA*Et)
        Mt = np.linalg.inv(Mt)
        Vt = np.dot(Id-Et, J[:,t])
        Vt = np.dot(Mt, Vt)
        Mag[t] = np.sum(Vt)

    # Return in original shape
    R1 = R1.reshape(n)
    return Mag.reshape(n[1:])


def signal_ss(R1, S0, TR, FA, v=None, PSw=np.inf, R10=None)->np.ndarray:
    """Signal of a spoiled gradient echo sequence applied in steady state.

    Args:
        R1 (array-like): Longitudinal relaxation rates in 1/sec. For a tissue with n compartments, the first dimension of R1 must be n. For a tissue with a single compartment, R1 can have any shape.
        S0 (float): Signal scaling factor (arbitrary units).
        TR (float): Repetition time, or time between successive selective excitations, in sec. 
        FA (float): Flip angle in degrees.
        v (array-like, optional): volume fractions of each compartment. If v is not provided, the tissue will be treated as one-compartmental. If v is provided, the length must be same as the first dimension of R1 and values must add up to 1. Defaults to None.
        PSw (array-like, optional): Water permeability-surface area products through the interfaces between the compartments, in units of mL/sec/mL. With PSw=np.inf (default), water exchange is in the fast-exchange limit. With PSw=0, there is no water exchange between the compartments. For any intermediate level of water exchange, PSw must be a nxn array, where n is the number of compartments, and PSw[j,i] is the permeability for water moving from compartment i into j. The diagonal elements PSw[i,i] quantify the flow of water from compartment i to outside. Defaults to np.inf.
        R10 (float, optional): R1-value where S0 is defined. If not provided, S0 is the scaling factor corresponding to infinite R10. Defaults to None.

    Returns:
        np.ndarray: Signal in the same units as S0.
    """
    if R10 is None:
        Sinf = S0
    else:
        Sinf = S0/signal_ss(R10, 1, TR, FA, v=v, PSw=PSw)
    if v is None:
        return _signal_ss(R1, Sinf, TR, FA)
    if np.size(v) != np.shape(R1)[0]:
        raise ValueError('v must have the same length as the first dimension of R1.')
    if np.isscalar(PSw):
        if PSw==np.inf:
            return _signal_ss_fex(v, R1, Sinf, TR, FA)
        elif PSw==0:
            return _signal_ss_nex(v, R1, Sinf, TR, FA)
    else:
        if np.ndim(PSw) != 2:
            raise ValueError("For intermediate water exchange, PSw must be a square array")
        if np.shape(PSw)[0] != np.size(v):
            raise ValueError("Dimensions of PSw and v do not match up.")
        return _signal_ss_aex(PSw, v, R1, Sinf, TR, FA)


def conc_ss(S, TR:float, FA:float, T10:float, r1=0.005, n0=1)->np.ndarray:
    """Concentration of a spoiled gradient echo sequence applied in steady state.

    Args:
        S (array-like): Signal in arbitrary units.
        TR (float): Repetition time, or time between successive selective excitations, in sec.
        FA (float): Flip angle in degrees.
        T10 (float): baseline T1 value in sec.
        r1 (float, optional): Longitudinal relaxivity in Hz/M. Defaults to 0.005.
        n0 (int, optional): Baseline length. Defaults to 1.

    Returns:
        np.ndarray: Concentration in M, same length as S.
    """
    # S = Sinf * (1-exp(-TR*R1)) / (1-cFA*exp(-TR*R1))
    # Sb = Sinf * (1-exp(-TR*R10)) / (1-cFA*exp(-TR*R10))
    # Sn = (1-exp(-TR*R1)) / (1-cFA*exp(-TR*R1))
    # Sn * (1-cFA*exp(-TR*R1)) = 1-exp(-TR*R1)
    # exp(-TR*R1) - Sn *cFA*exp(-TR*R1) = 1-Sn
    # (1-Sn*cFA) * exp(-TR*R1) = 1-Sn
    Sb = np.mean(S[:n0])
    E0 = math.exp(-TR/T10)
    c = math.cos(FA*math.pi/180)
    Sn = (S/Sb)*(1-E0)/(1-c*E0)	        # normalized signal
    # Replace any Nan values by interpolating between nearest neighbours
    outrange = Sn >= 1
    if np.sum(outrange) > 0:
        inrange = Sn < 1
        x = np.arange(Sn.size)
        Sn[outrange] = np.interp(x[outrange], x[inrange], Sn[inrange])
    R1 = -np.log((1-Sn)/(1-c*Sn))/TR	# relaxation rate in 1/msec
    return (R1 - 1/T10)/r1 



def signal_sr(R1, S0:float, TR:float, FA:float, TC:float, TP=0.0, R10=None)->np.ndarray:
    """Signal model for a saturation-recovery sequence with a FLASH readout.

    Args:
        R1 (array-like): Longitudinal relaxation rate in 1/sec.
        S0 (float): Signal scaling factor (arbitrary units).
        TR (float): Repetition time, or time between successive selective excitations, in sec. 
        FA (array-like): Flip angle in degrees.
        TC (float): Time (sec) between the saturation pulse and the acquisition of the k-space center.
        TP (float, optional): Time (sec) between the saturation pre-pulse and the first readout pulse. Defaults to 0.

    Raises:
        ValueError: If TP is larger than TC.

    Returns:
        np.ndarray: Signal in arbitrary units, of the same length as R1.
    """
    if R10 is None:
        Sinf = S0
    else:
        Sinf = S0/signal_sr(R10, 1, TR, FA, TC, TP)
    if TP > TC:
        msg = 'Incorrect sequence parameters.'
        msg += 'Tsat must be smaller than TC.'
        raise ValueError(msg)
    cFA = np.cos(np.pi*FA/180)
    T1_app = TR/(TR*R1-np.log(cFA))

    ER = np.exp(-TR*R1)
    E_sat = np.exp(-TP*R1)
    E_center = np.exp(-(TC-TP)/T1_app)

    S_sat = Sinf * (1-E_sat)
    S_ss = Sinf * (1-ER)/(1-cFA*ER)

    return S_ss*(1-E_center) + S_sat*E_center


def signal_sr_fex(v, R1, S0:float, TR:float, FA:float, TC:float, TP=0.0):
    """Signal of a spoiled gradient echo sequence applied after satration preparation, for a multi-compartmental tissue with fast water exchange.

    Args:
        v (array-like): Volume fractions of the compartments. v is a numpy array with size nc, where nc is the number of compartments.
        R1 (np.ndarray): Longitudinal relaxation rates of the compartments in 1/sec. R1 is a numpy array with dimensions (nc,) or (nc,nt), where nc is the number of compartments and nt is tne number of time points.
        S0 (float): Signal scaling factor (arbitrary units).
        TR (float): Repetition time, or time between successive selective excitations, in sec. 
        FA (array-like): Flip angle in degrees.
        TC (float): Time (sec) between the saturation pulse and the acquisition of the k-space center.
        TP (float, optional): Time (sec) between the saturation pre-pulse and the first readout pulse. Defaults to 0.

    Returns:
        np.ndarray: Signal in arbitrary units. If R1 is two-dimensional, this returns an array with a signal value for each time point.
    """
    if np.size(R1) == np.size(v):
        R1 = np.sum(np.multiply(v,R1))
        return signal_sr(R1, S0, TR, FA, TC, TP)
    nc, nt = R1.shape
    R1fex = np.zeros(nt)
    for c in range(nc):
        R1fex += v[c]*R1[c,:]
    return signal_sr(R1fex, S0, TR, FA, TC, TP)


def signal_sr_nex(v, R1, S0:float, TR:float, FA:float, TC:float, TP=0.0):
    """Signal of a spoiled gradient echo sequence applied with saturation preparation, for a multi-compartmental tissue without water exchange.

    Args:
        v (array-like): Volume fractions of the compartments. v is a numpy array with size nc, where nc is the number of compartments.
        R1 (np.ndarray): Longitudinal relaxation rates of the compartments in 1/sec. R1 is a numpy array with dimensions (nc,) or (nc,nt), where nc is the number of compartments and nt is tne number of time points.
        S0 (float): Signal scaling factor (arbitrary units).
        TR (float): Repetition time, or time between successive selective excitations, in sec. 
        FA (array-like): Flip angle in degrees.
        TC (float): Time (sec) between the saturation pulse and the acquisition of the k-space center.
        TP (float, optional): Time (sec) between the saturation pre-pulse and the first readout pulse. Defaults to 0.

    Returns:
        np.ndarray: Signal in arbitrary units. If R1 is two-dimensional, this returns an array with a signal value for each time point.
    """
    if np.size(R1) == np.size(v):
        S = signal_sr(R1, S0, TR, FA, TC, TP)
        return np.sum(np.multiply(v,S)) 
    nc, nt = R1.shape
    S = np.zeros(nt)
    for c in range(nc):
        S += v[c]*signal_sr(R1[c,:], S0, TR, FA, TC, TP)
    return S


def signal_er(R1, S0:float, TR:float, FA:float, TC:float)->np.ndarray:
    """Signal model for a FLASH readout, starting from equilibrium (i.e. no preparation pulse).

    Args:
        R1 (array-like): Longitudinal relaxation rate in 1/sec.
        S0 (float): Signal scaling factor (arbitrary units).
        TR (float): Repetition time, or time between successive selective excitations, in sec. 
        FA (array-like): Flip angle in degrees.
        TC (float): Time (sec) between the saturation pulse and the acquisition of the k-space center.

    Returns:
        np.ndarray: Signal in arbitrary units, of the same length as R1.
    """
    #TI is the residence time in the slab
    cFA = np.cos(np.pi*FA/180)
    R1_app = R1 - np.log(cFA)/TR

    ER = np.exp(-TR*R1)
    EI = np.exp(-TC*R1_app)

    S_ss = S0 * (1-ER)/(1-cFA*ER)

    return S_ss*(1-EI) + S0*EI


def signal_src(R1, S0, TC, R10=None):
    """Signal model for a saturation-recovery with a center-encoded readout.

    This can also be used with other encoding schemens whenever the effect of the readout pulses can be ignored, such as for fast flowing magnetization in arterial blood.

    Args:
        R1 (array-like): Longitudinal relaxation rate in 1/sec.
        S0 (float): Signal scaling factor (arbitrary units).
        TC (float): Time (sec) between the saturation pulse and the acquisition of the k-space center.

    Returns:
        np.ndarray: Signal in arbitrary units, of the same length as R1.
    """
    if R10 is None:
        Sinf = S0
    else:
        Sinf = S0/signal_src(R10, 1, TC)
    E = np.exp(-TC*R1)
    return Sinf * (1-E)


def conc_src(S, TC:float, T10:float, r1=0.005, n0=1)->np.ndarray:
    """Concentration of a saturation-recovery sequence with a center-encoded readout.

    Args:
        S (array-like): Signal in arbitrary units.
        TC (float): Time (sec) between the saturation pulse and the acquisition of the k-space center.
        T10 (float): baseline T1 value in sec.
        r1 (float, optional): Longitudinal relaxivity in Hz/M. Defaults to 0.005.
        n0 (int, optional): Baseline length. Defaults to 1.

    Returns:
        np.ndarray: Concentration in M, same length as S.

    Example:

        We generate some signals from ground-truth concentrations, then reconstruct the concentrations and check against the ground truth:

    .. plot::
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import dcmri as dc

        First define some constants:

        >>> T10 = 1         # sec
        >>> TC = 0.2        # sec
        >>> r1 = 0.005      # Hz/M

        Generate ground truth concentrations and signal data:

        >>> t = np.arange(0, 5*60, 0.1)     # sec
        >>> C = 0.003*(1-np.exp(-t/60))     # M
        >>> R1 = 1/T10 + r1*C               # Hz
        >>> S = dc.signal_src(R1, 100, TC)  # au

        Reconstruct the concentrations from the signal data:

        >>> Crec = dc.conc_src(S, TC, T10, r1)

        Check results by plotting ground truth against reconstruction:

        >>> plt.plot(t/60, 1000*C, 'ro', label='Ground truth')
        >>> plt.plot(t/60, 1000*Crec, 'b-', label='Reconstructed')
        >>> plt.title('SRC signal inverse')
        >>> plt.xlabel('Time (min)')
        >>> plt.ylabel('Concentration (mM)')
        >>> plt.legend()
        >>> plt.show()

    """
    # S = S0*(1-exp(-TC*R1))
    # S/Sb = (1-exp(-TC*R1))/(1-exp(-TC*R10))
    # (1-exp(-TC*R10))*S/Sb = 1-exp(-TC*R1)
    # 1-(1-exp(-TC*R10))*S/Sb = exp(-TC*R1)
    # ln(1-(1-exp(-TC*R10))*S/Sb) = -TC*R1
    # -ln(1-(1-exp(-TC*R10))*S/Sb)/TC = R1
    Sb = np.mean(S[:n0])
    E = math.exp(-TC/T10)
    R1 = -np.log(1-(1-E)*S/Sb)/TC	
    return (R1 - 1/T10)/r1 


def signal_lin(R1, S0:float)->np.ndarray:
    """Signal for any sequence operating in the linear regime.

    Args:
        R1 (array-like): Longitudinal relaxation rate in 1/sec.
        S0 (float): Signal scaling factor (arbitrary units).

    Returns:
        np.ndarray: Signal in arbitrary units, of the same length as R1.
    """
    return S0 * R1


def conc_lin(S, T10, r1=0.005, n0=1):
    """Concentration for any sequence operating in the linear regime.

    Args:
        S (array-like): Signal in arbitrary units.
        T10 (float): baseline T1 value in sec.
        r1 (float, optional): Longitudinal relaxivity in Hz/M. Defaults to 0.005.
        n0 (int, optional): Baseline length. Defaults to 1.

    Returns:
        np.ndarray: Concentration in M, same length as S.
    """
    Sb = np.mean(S[:n0])
    R10 = 1/T10
    R1 = R10*S/Sb	#relaxation rate in 1/msec
    return (R1 - R10)/r1 


def sample(t, tp, Sp, dt=None)->np.ndarray: 
    """Sample a signal at given time points.

    Args:
        t (array-like): The time points at which to evaluate the signal.
        tp (array-like): the time points of the signal to be sampled.
        Sp (array-like): the values of the signal to be sampled. Values that are outside of the range are set to zero.
        dt (float, optional): sampling interval. If this is not provided, linear interpolation between the data points is used.  Defaults to None.

    Returns:
        np.ndarray: Signals sampled at times t.
    """
    if dt is None:#
        return np.interp(t, tp, Sp, left=0, right=0)
    Ss = np.zeros(len(t)) 
    for k, tk in enumerate(t):
        data = Sp[(tp >= tk) & (tp < tk+dt)]
        if data.size > 0:
            Ss[k] = np.mean(data)
    return Ss 


def add_noise(signal, sdev:float)->np.ndarray:
    """Add noise to an MRI magnitude signal.

    Args:
        signal (array-like): Signal values.
        sdev (float): Standard deviation of the noise.

    Returns:
        np.ndarray: signal with noise added.
    """
    noise_x = np.random.normal(0, sdev, signal.size)
    noise_y = np.random.normal(0, sdev, signal.size)
    signal = np.sqrt((signal+noise_x)**2 + noise_y**2)
    return signal


import numpy as np
import dcmri as dc
from scipy.integrate import cumulative_trapezoid

def test__conc_u():
    n = 10
    Ta = 10
    Fp = 2
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    C = dc.conc_tissue(ca, Fp, t=t, kinetics='U')
    C0 = Fp*cumulative_trapezoid(ca, t, initial=0)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.01

def test__flux_u():
    n = 10
    Ta = 10
    Fp = 2
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    J = dc.flux_tissue(ca, Fp, t=t, kinetics='U')
    J0 = np.zeros(len(t))
    assert np.linalg.norm(J-J0) < 0.01

def test__conc_1c():
    n = 10
    Ta = 10
    Fp = 2
    v = 0.1
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    C = dc.conc_tissue(ca, Fp, v, t=t, kinetics='FX')
    C0 = Fp*dc.biexpconv(Ta, v/Fp, t)*v/Fp
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.01
    C = dc.conc_tissue(ca, 0, v, t=t, kinetics='FX')
    assert np.linalg.norm(C) == 0

def test__flux_1c():
    n = 10
    Ta = 10
    Fp = 2
    v = 0.1
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    J = dc.flux_tissue(ca, Fp, v, t=t, kinetics='FX')
    J0 = Fp*dc.biexpconv(Ta, v/Fp, t)
    assert np.linalg.norm(J-J0)/np.linalg.norm(J0) < 0.01
    J = dc.flux_tissue(ca, 0, v, t=t, kinetics='FX')
    assert np.linalg.norm(J) == 0
    

def test__flux_wv():
    n = 10
    Ta = 10
    Ktrans = 2
    kep = 5
    ve = Ktrans/kep
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    J = dc.flux_tissue(ca, Ktrans, ve, t=t, kinetics='WV')
    J0 = Ktrans*dc.biexpconv(Ta, 1/kep, t)
    assert np.linalg.norm(J[0,1,:]-J0)/np.linalg.norm(J0) < 0.01
    J = dc.flux_tissue(ca, 0, ve, t=t, kinetics='WV')
    assert np.linalg.norm(J[0,1,:]) == 0
    

def test__conc_hfu():
    n = 10
    Ta = 10
    vp = 0.3
    Ktrans = 2
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    C = dc.conc_tissue(ca, vp, Ktrans, t=t, kinetics='HFU', sum=False)
    C0 = vp*ca
    C1 = Ktrans*dc.conc_trap(ca, t)
    assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 0.01
    assert np.linalg.norm(C[1,:]-C1)/np.linalg.norm(C1) < 0.01
    C = dc.conc_tissue(ca, vp, Ktrans, t=t, kinetics='HFU', sum=True)
    C0 = C0+C1
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.01

def test__flux_hfu():
    n = 10
    Ta = 10
    vp = 0.3
    Ktrans = 2
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    J = dc.flux_tissue(ca, vp, Ktrans, t=t, kinetics='HFU')
    J0 = Ktrans*ca
    assert np.linalg.norm(J[1,0,:]-J0)/np.linalg.norm(J0) < 0.01

def test_conc_hf():
    n = 10
    Ta = 10
    vp = 0.3
    Ktrans = 2
    ve = 0.2
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    C = dc.conc_tissue(ca, vp, Ktrans, ve, t=t, kinetics='HF', sum=False)
    C0 = vp*ca
    C1 = Ktrans*dc.biexpconv(Ta, ve/Ktrans, t)*ve/Ktrans
    assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 0.01
    assert np.linalg.norm(C[1,:]-C1)/np.linalg.norm(C1) < 0.01
    C = dc.conc_tissue(ca, vp, Ktrans, ve, t=t, kinetics='HF', sum=True)
    C0 = C0+C1
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.01
    C = dc.conc_tissue(ca, vp, Ktrans, 0, t=t, kinetics='HF', sum=False)
    C0 = vp*ca
    C1 = np.zeros(len(ca))
    assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 0.01
    assert np.linalg.norm(C[1,:]-C1) == 0

def test__flux_hf():
    n = 10
    Ta = 10
    vp = 0.3
    Ktrans = 2
    kep = 5
    ve = Ktrans/kep
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    J = dc.flux_tissue(ca, vp, Ktrans, ve, t=t, kinetics='HF')
    J0 = kep*Ktrans*dc.biexpconv(Ta, 1/kep, t)/kep
    assert np.linalg.norm(J[0,1,:]-J0)/np.linalg.norm(J0) < 0.01

def test__conc_2cu():
    n = 10
    Ta = 10
    vp = 0.3
    PS = 2
    Fp = 4
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    C = dc.conc_tissue(ca, Fp, vp, PS, t=t, sum=False, kinetics='2CU')
    Tp = vp/(Fp+PS)
    C0 = Tp*Fp*dc.biexpconv(Ta, Tp, t)
    assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 0.01
    Cs = dc.conc_tissue(ca, Fp, vp, PS, t=t, sum=True, kinetics='2CU')
    C0 = C0+C[1,:]
    assert np.linalg.norm(Cs-C0)/np.linalg.norm(C0) < 0.01
    Cs = dc.conc_tissue(ca, Fp, 0, PS, t=t, sum=True, kinetics='2CU')
    Ktrans = Fp*PS/(Fp+PS)
    C0 = dc.conc_tissue(ca, Ktrans, t=t, kinetics='U')
    assert np.linalg.norm(Cs-C0)/np.linalg.norm(C0) < 0.01
    Cs = dc.conc_tissue(ca, 0, vp, 0, t=t, sum=True, kinetics='2CU')
    assert np.linalg.norm(Cs)==0

def test__flux_2cu():
    n = 10
    Ta = 10
    vp = 0.3
    PS = 2
    Fp = 4
    t = np.linspace(0, 20, n)
    ca = np.exp(-t/Ta)/Ta
    J = dc.flux_tissue(ca, Fp, vp, PS, t=t, kinetics='2CU')
    Tp = vp/(Fp+PS)
    J0 = Tp*Fp*dc.biexpconv(Ta, Tp, t)*Fp/vp
    assert np.linalg.norm(J[0,0,:]-J0)/np.linalg.norm(J0) < 0.01
    J = dc.flux_tissue(ca, Fp, 0, PS, t=t, kinetics='2CU')
    assert np.linalg.norm(J[0,0,:]-Fp*ca)/np.linalg.norm(Fp*ca) < 0.01

def test__conc_2cx():
    # Compare against general 2comp solution
    t = np.linspace(0, 20, 10)
    J = np.ones((2,len(t)))
    J[1,:] = 0 # No inlet in 2nd compartment
    T = [6,12]
    E = 0.2
    Emat = [[1-E,1],[E,0]]
    Fp = 1/60 # mL/sec/mL
    ca = J[0,:]/Fp
    PS = Fp*E/(1-E)
    ve = T[1]*PS
    vp = T[0]*(Fp+PS)
    C0 = dc.conc_tissue(ca, Fp, vp, PS, ve, t=t, sum=False, kinetics='2CX')
    C = dc.conc_2comp(J, T, Emat, t)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C) < 1e-3
    Cs = dc.conc_tissue(ca, Fp, vp, PS, ve, t=t, sum=True, kinetics='2CX')
    C0 = np.sum(C,axis=0)
    assert np.linalg.norm(Cs-C0)/np.linalg.norm(C0) < 0.01
    Cs = dc.conc_tissue(ca, 0, vp, 0, ve, t=t, sum=True, kinetics='2CX')
    assert np.linalg.norm(Cs) == 0
    Cs = dc.conc_tissue(ca, 0, vp, 0, ve, t=t, sum=False, kinetics='2CX')
    assert np.linalg.norm(Cs) == 0
    Cs = dc.conc_tissue(ca, Fp, vp, 0, ve, t=t, sum=True, kinetics='2CX')
    C0 = dc.conc_tissue(ca, Fp, vp, t=t, kinetics='NX')
    assert np.linalg.norm(Cs-C0)/np.linalg.norm(C) < 1e-3
    C = dc.conc_tissue(ca, Fp, vp, 0, ve, t=t, sum=False, kinetics='2CX')
    assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 1e-3
    
def test__flux_2cx():
    # Compare against general 2comp solution
    t = np.linspace(0, 20, 10)
    J = np.ones((2,len(t)))
    J[1,:] = 0 # No inlet in 2nd compartment
    T = [6,12]
    E = 0.2
    Emat = [[1-E,1],[E,0]]
    Fp = 1/60 # mL/sec/mL
    ca = J[0,:]/Fp
    PS = Fp*E/(1-E)
    ve = T[1]*PS
    vp = T[0]*(Fp+PS)
    Jo0 = dc.flux_tissue(ca, Fp, vp, PS, ve, t=t, kinetics='2CX') 
    Jo = dc.flux_2comp(J, T, Emat, t)
    assert np.linalg.norm(Jo-Jo0)/np.linalg.norm(Jo) < 1e-3
    Jo0 = dc.flux_tissue(ca, 0, vp, 0, ve, t=t, kinetics='2CX') 
    assert np.linalg.norm(Jo0) == 0
    Jo0 = dc.flux_tissue(ca, Fp, vp, 0, ve, t=t, kinetics='2CX') 
    Jo = dc.flux_tissue(ca, Fp, vp, t=t, kinetics='NX')
    assert np.linalg.norm(Jo-Jo0[0,0,:])/np.linalg.norm(Jo) < 1e-3


def test__conc_2cf():
    # Compare against general 2comp solution
    t = np.linspace(0, 20, 10)
    J = np.ones((2,len(t)))
    J[1,:] = 0 # No inlet in 2nd compartment
    T = [6,12]
    E = 0.2
    Fp = 1/60 # mL/sec/mL
    ca = J[0,:]/Fp
    PS = Fp*E/(1-E)
    Te = T[1]
    vp = T[0]*(Fp+PS)
    C0 = dc.conc_tissue(ca, Fp, vp, PS, Te, t=t, sum=False, kinetics='2CF') 
    Emat = [
        [1-E, 0],
        [E  , 1]]
    C = dc.conc_2comp(J, T, Emat, t)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C) < 1e-2
    Cs = dc.conc_tissue(ca, Fp, vp, PS, Te, t=t, sum=True, kinetics='2CF')
    C0 = np.sum(C,axis=0)
    assert np.linalg.norm(Cs-C0)/np.linalg.norm(C0) < 0.01
    C = dc.conc_tissue(ca, Fp, vp, 0, Te, t=t, sum=False, kinetics='2CF')
    C0 = dc.conc_comp(J[0,:], vp/Fp, t)
    assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 1e-12
    C = dc.conc_tissue(ca, Fp, 0, PS, Te, t=t, sum=False, kinetics='2CF')
    assert np.linalg.norm(C[0,:]) == 0
    Cs = dc.conc_tissue(ca, 0, vp, 0, Te, t=t, sum=True, kinetics='2CF')
    assert np.linalg.norm(Cs) == 0
    Cs = dc.conc_tissue(ca, 0, vp, 0, Te, t=t, sum=False, kinetics='2CF')
    assert np.linalg.norm(Cs) == 0


def test__flux_2cf():
    # Compare against general 2comp solution
    t = np.linspace(0, 20, 10)
    J = np.ones((2,len(t)))
    J[1,:] = 0 # No inlet in 2nd compartment
    T = [6,12]
    E = 0.2
    Fp = 1/60 # mL/sec/mL
    ca = J[0,:]/Fp
    PS = Fp*E/(1-E)
    Te = T[1]
    vp = T[0]*(Fp+PS)
    Jo0 = dc.flux_tissue(ca, Fp, vp, PS, Te, t=t, kinetics='2CF') 
    Emat = [
        [1-E, 0],
        [E  , 1]]
    Jo = dc.flux_2comp(J, T, Emat, t)
    assert np.linalg.norm(Jo-Jo0)/np.linalg.norm(Jo) < 1e-3
    Jo0 = dc.flux_tissue(ca, 0, vp, 0, Te, t=t, kinetics='2CF') 
    assert np.linalg.norm(Jo0) == 0




if __name__ == "__main__":

    test__conc_u()
    test__flux_u()

    test__conc_1c()
    test__flux_1c()

    test__flux_wv()

    test__conc_hfu()
    test__flux_hfu()

    test_conc_hf()
    test__flux_hf()

    test__conc_2cu()
    test__flux_2cu()

    test__conc_2cx()
    test__flux_2cx()

    test__conc_2cf()
    test__flux_2cf()

    print('All pkmods tests passing!')


"""PK models built from PK blocks defined in dcmri.pk"""

import numpy as np
import dcmri.pk as pk
import dcmri.utils as utils


def conc_tissue(ca:np.ndarray, *params, t=None, dt=1.0, kinetics='2CX', sum=True)->np.ndarray:
    """Tissue concentration in a 2-site exchange tissue.

    Args:
        ca (array-like): concentration in the arterial input.
        params (tuple): free model parameters.
        t (array_like, optional): the time points in sec of the input function *ca*. If *t* is not provided, the time points are assumed to be uniformly spaced with spacing *dt*. Defaults to None.
        dt (float, optional): spacing in seconds between time points for uniformly spaced time points. This parameter is ignored if *t* is explicity provided. Defaults to 1.0.
        kinetics (str, optional): Kinetics of the tissue, either 'U', 'NX', 'FX', 'WV', 'HFU', 'HF', '2CU', '2CX', '2CF' - see below for detail. Defaults to '2CX'. 
        sum (bool, optional): For two-compartment tissues, set to True to return the total tissue concentration. Defaults to True.

    Returns:
        numpy.ndarray: If sum=True, this is a 1D array with the total concentration at each time point. If sum=False this is the concentration in each compartment, and at each time point, as a 2D array with dimensions *(2,k)*, where *k* is the number of time points in *ca*. The concentration is returned in units of M.

    Notes:
        Currently implemented kinetic models are: 
        
        - 'U': uptake tissue. params = Fp
        - 'NX': no tracer exchange tissue. params = (Fp, vp, )
        - 'FX': fast tracer exchange tissue. params = (Fp, v, )
        - 'WV': weakly vascularized tissue - also known as *Tofts model*. params = (Ktrans, ve, )
        - 'HFU': high-flow uptake tissue - also known as *Patlak model*. params = (vp, PS, )
        - 'HF': high-flow tissue - also known as *extended Tofts model*, *extended Patlak model* or *general kinetic model*. Params = (vp, PS, ve, )
        - '2CU': two-compartment uptake tissue. params = (Fp, vp, PS, )
        - '2CX': two-compartment exchange tissue. params = (Fp, vp, PS, ve, )
        - '2CF': two-compartment filtration tissue. params = (Fp, vp, PS, Te, )

        The model parameters are:

        - **Fp** (mL/sec/mL): Plasma flow.
        - **PS** (mL/sec/mL): permeability surface-area product.
        - **Ktrans** (mL/mL): volume transfer constant = Fp*PS/(FP+PS) 
        - **vp** (mL/mL): plasma volume fraction.
        - **ve** (mL/mL): extracellular, extravascular volume fraction.
        - **v** (mL/mL): extracellular volume fraction = vp+ve.
        - **Te** (sec): extracellular, extravascular mean transit time.

    Example:

        Plot concentration in cortex and medulla for typical values:

    .. plot::
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import dcmri as dc

        Generate a population-average input function:

        >>> t = np.arange(0, 300, 1.5)
        >>> ca = dc.aif_parker(t, BAT=20)

        Define some tissue parameters: 

        >>> Fp, vp, PS, ve = 0.01, 0.05, 0.005, 0.4
        >>> Ktrans = PS*Fp/(PS+Fp)

        Set up a plot to show concentrations:

        >>> fig, (ax0, ax1) = plt.subplots(1,2,figsize=(12,5))
        
        Generate plasma and extravascular tissue concentrations with the 2CX model and add to the plot:

        >>> C = dc.conc_tissue(ca, Fp, vp, PS, ve, t=t, sum=False, kinetics='2CX')
        >>> ax0.set_title('2-compartment exchange model')
        >>> ax0.plot(t/60, 1000*C[0,:], linestyle='--', linewidth=3.0, color='darkred', label='Plasma')
        >>> ax0.plot(t/60, 1000*C[1,:], linestyle='--', linewidth=3.0, color='darkblue', label='Extravascular, extracellular space')
        >>> ax0.plot(t/60, 1000*(C[0,:]+C[1,:]), linestyle='-', linewidth=3.0, color='grey', label='Tissue')
        >>> ax0.set_xlabel('Time (min)')
        >>> ax0.set_ylabel('Tissue concentration (mM)')
        >>> ax0.legend()

        Generate plasma and extravascular tissue concentrations with the WV model and add to the plot: 

        >>> C = dc.conc_tissue(ca, Ktrans, ve, t=t, sum=False, kinetics='WV')
        >>> ax1.set_title('Weakly vascularised model')
        >>> ax1.plot(t/60, 1000*C[0,:], linestyle='--', linewidth=3.0, color='darkred', label='Plasma (WV)')
        >>> ax1.plot(t/60, 1000*C[1,:], linestyle='--', linewidth=3.0, color='darkblue', label='Extravascular, extracellular space (WV)')
        >>> ax1.plot(t/60, 1000*(C[0,:]+C[1,:]), linestyle='-', linewidth=3.0, color='grey', label='Tissue (WV)')
        >>> ax1.set_xlabel('Time (min)')
        >>> ax1.set_ylabel('Tissue concentration (mM)')
        >>> ax1.legend()
        >>> plt.show()

    """
    if kinetics=='U':
        return _conc_u(ca, *params, t=t, dt=dt)
    elif kinetics=='NX':
        return _conc_1c(ca, *params, t=t, dt=dt)
    elif kinetics=='FX':
        return _conc_1c(ca, *params, t=t, dt=dt)
    elif kinetics=='WV':
        return _conc_wv(ca, *params, t=t, dt=dt, sum=sum)
    elif kinetics=='HFU':
        return _conc_hfu(ca, *params, t=t, dt=dt, sum=sum)
    elif kinetics=='HF':
        return _conc_hf(ca, *params, t=t, dt=dt, sum=sum)
    elif kinetics=='2CU':
        return _conc_2cu(ca, *params, t=t, dt=dt, sum=sum)
    elif kinetics=='2CX':
        return _conc_2cx(ca, *params, t=t, dt=dt, sum=sum)
    elif kinetics=='2CF':
        return _conc_2cf(ca, *params, t=t, dt=dt, sum=sum)
    else:
        raise ValueError('Kinetic model ' + kinetics + ' is not currently implemented.')



def flux_tissue(ca:np.ndarray, *params, t=None, dt=1.0, kinetics='2CX')->np.ndarray:
    """Indicator out of a 2-site exchange tissue.

    Args:
        ca (array-like): concentration in the arterial input.
        params (tuple): free model parameters.
        t (array_like, optional): the time points in sec of the input function *ca*. If *t* is not provided, the time points are assumed to be uniformly spaced with spacing *dt*. Defaults to None.
        dt (float, optional): spacing in seconds between time points for uniformly spaced time points. This parameter is ignored if *t* is explicity provided. Defaults to 1.0.
        kinetics (str, optional): Kinetics of the tissue, either 'U', 'NX', 'FX', 'WV', 'HFU', 'HF', '2CU', '2CX', '2CF' - see below for detail. Defaults to '2CX'. 

    Returns:
        numpy.ndarray: For a one-compartmental tissue, outflux out of the compartment as a 1D array in units of mmol/sec/mL or M/sec. For a multi=compartmental tissue, outflux out of each compartment, and at each time point, as a 3D array with dimensions *(2,2,k)*, where *2* is the number of compartments and *k* is the number of time points in *J*. Encoding of the first two indices is the same as for *E*: *J[j,i,:]* is the flux from compartment *i* to *j*, and *J[i,i,:]* is the flux from *i* directly to the outside. The flux is returned in units of mmol/sec/mL or M/sec.

    Notes:
        Currently implemented kinetic models are: 
        
        - 'U': uptake tissue. params = Fp
        - 'NX': no tracer exchange tissue. params = (Fp, vp, )
        - 'FX': fast tracer exchange tissue. params = (Fp, v, )
        - 'WV': weakly vascularized tissue - also known as *Tofts model*. params = (Ktrans, ve, )
        - 'HFU': high-flow uptake tissue - also known as *Patlak model*. params = (vp, PS, )
        - 'HF': high-flow tissue - also known as *extended Tofts model*, *extended Patlak model* or *general kinetic model*. Params = (vp, PS, ve, )
        - '2CU': two-compartment uptake tissue. params = (Fp, vp, PS, )
        - '2CX': two-compartment exchange tissue. params = (Fp, vp, PS, ve, )
        - '2CF': two-compartment filtration tissue. params = (Fp, vp, PS, Te, )

        The model parameters are:

        - **Fp** (mL/sec/mL): Plasma flow.
        - **PS** (mL/sec/mL): permeability surface-area product.
        - **Ktrans** (mL/mL): volume transfer constant = Fp*PS/(FP+PS) 
        - **vp** (mL/mL): plasma volume fraction.
        - **ve** (mL/mL): extracellular, extravascular volume fraction.
        - **v** (mL/mL): extracellular volume fraction = vp+ve.
        - **Te** (sec): extracellular, extravascular mean transit time.
        
    """
    if kinetics=='U':
        return _flux_u(ca, *params, t=t, dt=dt)
    elif kinetics=='NX':
        return _flux_1c(ca, *params, t=t, dt=dt)
    elif kinetics=='FX':
        return _flux_1c(ca, *params, t=t, dt=dt)
    elif kinetics=='WV':
        return _flux_wv(ca, *params, t=t, dt=dt)
    elif kinetics=='HFU':
        return _flux_hfu(ca, *params, t=t, dt=dt)
    elif kinetics=='HF':
        return _flux_hf(ca, *params, t=t, dt=dt)
    elif kinetics=='2CU':
        return _flux_2cu(ca, *params, t=t, dt=dt)
    elif kinetics=='2CX':
        return _flux_2cx(ca, *params, t=t, dt=dt)
    elif kinetics=='2CF':
        return _flux_2cf(ca, *params, t=t, dt=dt)
    else:
        raise ValueError('Kinetic model ' + kinetics + ' is not currently implemented.')



def _conc_u(ca, Fp, t=None, dt=1.0):
    return pk.conc_trap(Fp*ca, t=t, dt=dt)

def _conc_1c(ca, Fp, v, t=None, dt=1.0):
    if Fp==0:
        return np.zeros(len(ca))
    return pk.conc_comp(Fp*ca, v/Fp, t=t, dt=dt)

def _conc_wv(ca, Ktrans, ve, t=None, dt=1.0, sum=True):
    Cp = np.zeros(np.size(ca))
    if Ktrans==0:
        Ce = np.zeros(np.size(ca))
    else:
        Ce = pk.conc_comp(Ktrans*ca, ve/Ktrans, t=t, dt=dt)
    if sum:
        return Cp+Ce
    else:
        return np.stack((Cp,Ce))

def _conc_hfu(ca, vp, Ktrans, t=None, dt=1.0, sum=True):
    Cp = vp*ca
    Ce = pk.conc_trap(Ktrans*ca, t=t, dt=dt)
    if sum:
        return Cp+Ce
    else:
        return np.stack((Cp,Ce))
    
def _conc_hf(ca, vp, Ktrans, ve, t=None, dt=1.0, sum=True):
    Cp = vp*ca
    if Ktrans==0:
        Ce = 0*ca
    else:
        Ce = pk.conc_comp(Ktrans*ca, ve/Ktrans, t=t, dt=dt)
    if sum:
        return Cp+Ce
    else:
        return np.stack((Cp,Ce))
    
def _conc_2cu(ca, Fp, vp, PS, t=None, dt=1.0, sum=True):
    if Fp+PS==0:
        return np.zeros((2,len(ca)))
    Tp = vp/(Fp+PS)
    Cp = pk.conc_comp(Fp*ca, Tp, t=t, dt=dt)
    if vp==0:
        Ktrans = Fp*PS/(Fp+PS)
        Ce = pk.conc_trap(Ktrans*ca, t=t, dt=dt)
    else:
        Ce = pk.conc_trap(PS*Cp/vp, t=t, dt=dt)
    if sum:
        return Cp+Ce
    else:
        return np.stack((Cp,Ce))
    
def _conc_2cx(ca, Fp, vp, PS, ve, t=None, dt=1.0, sum=True):
    if Fp+PS == 0:
        if sum:
            return np.zeros(len(ca))
        else:
            return np.zeros((2,len(ca)))
    if PS == 0:
        Cp = _conc_1c(ca, Fp, vp, t=t, dt=dt)
        Ce = np.zeros(len(ca))
        if sum:
            return Cp+Ce
        else:
            return np.stack((Cp,Ce))
    # Derive standard parameters
    Tp = vp/(Fp+PS)
    Te = ve/PS
    J = Fp*ca
    E = PS/(Fp+PS)
    # Build the system matrix K
    T = [Tp, Te]
    E = [
        [1-E, 1],
        [E,   0],
    ]
    Q, K, Qi = pk.K_2comp(T, E)
    # Initialize concentration-time array
    nc, nt = 2, len(J)
    t = utils.tarray(nt, t=t, dt=dt)
    Ei = np.empty((nc,nt))
    # Loop over the eigenvalues
    for d in [0,1]:
        # Calculate elements of diagonal matrix
        Ei[d,:] = pk.conc_comp(J, 1/K[d], t)
        # Right-multiply with inverse eigenvector matrix
        Ei[d,:] *= Qi[d,0]
    # Left-multiply with eigenvector matrix
    C = np.matmul(Q, Ei)
    if sum:
        return np.sum(C, axis=0)
    else:
        return C
    
def _conc_2cf(ca, Fp, vp, PS, Te, t=None, dt=1.0, sum=True):
    if Fp+PS == 0:
        if sum:
            return np.zeros(len(ca))
        else:
            return np.zeros((2,len(ca)))
    # Derive standard parameters
    Tp = vp/(Fp+PS)
    E = PS/(Fp+PS)
    J = Fp*ca
    T = [Tp, Te]
    # Solve the system explicitly
    t = utils.tarray(len(J), t=t, dt=dt)
    C0 = pk.conc_comp(J, T[0], t)
    if E==0:
        C1 = np.zeros(len(t))
    elif T[0]==0:
        J10 = E*J
        C1 = pk.conc_comp(J10, T[1], t)
    else:
        J10 = C0*E/T[0]
        C1 = pk.conc_comp(J10, T[1], t)
    if sum:
        return C0+C1
    else:
        return np.stack((C0, C1))





def _flux_u(ca, Fp, t=None, dt=1.0):
    return pk.flux(Fp*ca, kinetics='trap')

def _flux_1c(ca, Fp, v, t=None, dt=1.0):
    if Fp==0:
        return np.zeros(len(ca))
    return pk.flux(Fp*ca, v/Fp, t=t, dt=dt, kinetics='comp')

def _flux_wv(ca, Ktrans, ve, t=None, dt=1.0):
    J = np.zeros(((2,2,len(ca))))
    J[0,0,:] = np.nan
    J[1,0,:] = Ktrans*ca
    if Ktrans!=0:
        J[0,1,:] = pk.flux(Ktrans*ca, ve/Ktrans, t=t, dt=dt, kinetics='comp')
    return J

def _flux_hfu(ca, vp, Ktrans, t=None, dt=1.0):
    J = np.zeros(((2,2,len(ca))))
    J[0,0,:] = np.nan
    J[1,0,:] = Ktrans*ca
    return J

def _flux_hf(ca, vp, Ktrans, ve, t=None, dt=1.0):
    J = np.zeros(((2,2,len(ca))))
    J[0,0,:] = np.nan
    J[1,0,:] = Ktrans*ca
    if Ktrans==0:
        J[0,1,:] = 0*ca
    else:
        J[0,1,:] = pk.flux(Ktrans*ca, ve/Ktrans, t=t, dt=dt, kinetics='comp')
    return J

def _flux_2cu(ca, Fp, vp, PS, t=None, dt=1.0):
    C = _conc_2cu(ca, Fp, vp, PS, t=t, dt=dt, sum=False)
    J = np.zeros(((2,2,len(ca))))
    if vp==0:
        J[0,0,:] = Fp*ca
        Ktrans = Fp*PS/(Fp+PS)
        J[1,0,:] = Ktrans*ca
    else:
        J[0,0,:] = Fp*C[0,:]/vp
        J[1,0,:] = PS*C[0,:]/vp
    return J


def _flux_2cx(ca, Fp, vp, PS, ve, t=None, dt=1.0):
    if Fp+PS == 0:
        return np.zeros((2,2,len(ca)))
    if PS == 0:
        Jp = _flux_1c(ca, Fp, vp, t=t, dt=dt)
        J = np.zeros((2,2,len(ca)))
        J[0,0,:] = Jp
        return J
    C = _conc_2cx(ca, Fp, vp, PS, ve, t=t, dt=dt, sum=False)
    # Derive standard parameters
    Tp = vp/(Fp+PS)
    Te = ve/PS
    J = Fp*ca
    E = PS/(Fp+PS)
    # Build the system matrix K
    T = [Tp, Te]
    E = [
        [1-E, 1],
        [E,   0],
    ]
    return pk.J_ncomp(C, T, E)


def _flux_2cf(ca, Fp, vp, PS, Te, t=None, dt=1.0):
    if Fp+PS == 0:
        return np.zeros((2,2,len(ca)))
    # Derive standard parameters
    Tp = vp/(Fp+PS)
    E = PS/(Fp+PS)
    J = Fp*ca
    T = [Tp, Te]
    # Solve the system explicitly
    t = utils.tarray(len(J), t=t, dt=dt)
    Jo = np.zeros((2,2,len(t)))
    J0 = pk.flux(J, T[0], t=t, kinetics='comp')   
    J10 = E*J0
    Jo[1,0,:] = J10
    Jo[1,1,:] = pk.flux(J10, T[1], t=t, kinetics='comp')
    Jo[0,0,:] = (1-E)*J0
    return Jo


